"""
Tests for Twisted instrumentation
"""

import json
import sys

from unittest import mock

import rollbar

# access token for https://rollbar.com/rollbar/pyrollbar
TOKEN = '92c10f5616944b81a2e6f3c6493a0ec2'

# Twisted hasn't been fully ported to Python 3 yet, so don't test there.
ALLOWED_PYTHON_VERSION = not sys.version_info[0] == 3
try:
    from twisted.test import proto_helpers
    from twisted.trial import unittest
    from twisted.internet import protocol
    from twisted.python import log

    TWISTED_INSTALLED = True
except ImportError:
    TWISTED_INSTALLED = False

if ALLOWED_PYTHON_VERSION and TWISTED_INSTALLED:
    class SquareProtocol(protocol.Protocol):
        def dataReceived(self, data):
            try:
                number = int(data)
            except ValueError:
                log.err()
                self.transport.write('error')
            else:
                self.transport.write(str(number**2))

    class SquareFactory(protocol.Factory):
        protocol = SquareProtocol

    class TwistedTest(unittest.TestCase):
        def setUp(self):
            rollbar.init(TOKEN, 'twisted-test')
            factory = SquareFactory()
            self.proto = factory.buildProtocol(('127.0.0.1', 0))
            self.tr = proto_helpers.StringTransport()
            self.proto.makeConnection(self.tr)

        @mock.patch('rollbar.send_payload')
        def test_base_case(self, send_payload):
            self.proto.dataReceived('8')
            self.assertEqual(int(self.tr.value()), 64)

            self.assertFalse(send_payload.called)

        @mock.patch('rollbar.send_payload')
        def test_caught_exception(self, send_payload):
            self.proto.dataReceived('rollbar')
            self.assertEqual(self.tr.value(), "error")
            errors = self.flushLoggedErrors(ValueError)
            self.assertEqual(len(errors), 1)

            self.assertTrue(send_payload.called)
            payload = send_payload.call_args[0][0]
            data = payload['data']

            self.assertIn('body', data)
            self.assertEqual(data['body']['trace']['exception']['class'],
                             'ValueError')
            self.assertEqual(data['body']['trace']['exception']['message'],
                             "invalid literal for int() with base 10: 'rollbar'")

        # XXX not able to test uncaught exceptions for some reason
        # @mock.patch('rollbar.send_payload')
        # def test_uncaught_exception(self, send_payload):
        #     self.proto.dataReceived([8, 9])
        #     self.assertEqual(self.tr.value(), "error")
        #     errors = self.flushLoggedErrors(TypeError)
        #     self.assertEqual(len(errors), 1)
        #
        #     self.assertEqual(send_payload.called, True)
        #     payload = send_payload.call_args[0][0]
        #     data = payload['data']
        #
        #     self.assertIn('body', data)
        #     self.assertEqual(data['body']['trace']['exception']['class'],
        #                      'TypeError')
        #     self.assertEqual(data['body']['trace']['exception']['message'],
        #                      "int() argument must be a string or a number, not 'list'")



import inspect
import functools

import rollbar


class IntegrationBase:
    """
    Superclass for class integrations.
    """

    def __init__(self):
        if hasattr(self, '_integrate'):
            self._integrate()


class integrate:
    """
    Integrates functions and classes (derived from IntegrationBase) in the SDK.
    """

    def __init__(self, *, framework_name='unknown'):
        self._framework_name = framework_name

    def __call__(self, obj):
        if inspect.isclass(obj):
            obj._integrate = self._register_hook
            return obj
        else:
            return self._insert_hook(obj)

    def _register_hook(self):
        def _hook(request, data):
            data['framework'] = self._framework_name

        rollbar.BASE_DATA_HOOK = _hook

    def _insert_hook(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self._register_hook()
            return func(*args, **kwargs)

        return wrapper



"""
Hooks for integrating with the python logging framework.

Usage:
    import logging
    from rollbar.logger import RollbarHandler

    rollbar.init('ACCESS_TOKEN', 'ENVIRONMENT')

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    # report ERROR and above to Rollbar
    rollbar_handler = RollbarHandler()
    rollbar_handler.setLevel(logging.ERROR)

    # attach the handlers to the root logger
    logger.addHandler(rollbar_handler)

"""
import logging
import threading

from logging.config import ConvertingDict, ConvertingList, ConvertingTuple

import rollbar

# hack to fix backward compatibility in Python3
try:
    from logging import _checkLevel
except ImportError:
    _checkLevel = lambda lvl: lvl


def resolve_logging_types(obj):
    if isinstance(obj, (dict, ConvertingDict)):
        return {k: resolve_logging_types(v) for k, v in obj.items()}
    elif isinstance(obj, (list, ConvertingList)):
        return [resolve_logging_types(i) for i in obj]
    elif isinstance(obj, (tuple, ConvertingTuple)):
        return tuple(resolve_logging_types(i) for i in obj)

    return obj


class RollbarHandler(logging.Handler):
    SUPPORTED_LEVELS = set(('debug', 'info', 'warning', 'error', 'critical'))

    _history = threading.local()

    def __init__(self,
                 access_token=None,
                 environment=None,
                 level=logging.INFO,
                 history_size=10,
                 history_level=logging.DEBUG,
                 **kw):

        logging.Handler.__init__(self)

        if access_token is not None:
            rollbar.init(
                access_token, environment, **resolve_logging_types(kw))

        self.notify_level = _checkLevel(level)

        self.history_size = history_size
        if history_size > 0:
            self._history.records = []

        self.setHistoryLevel(history_level)

    def setLevel(self, level):
        """
        Override so we set the effective level for which
        log records we notify Rollbar about instead of which
        records we save to the history.
        """
        self.notify_level = _checkLevel(level)

    def setHistoryLevel(self, level):
        """
        Use this method to determine which records we record history
        for. Use setLevel() to determine which level we report records
        to Rollbar for.
        """
        logging.Handler.setLevel(self, level)

    def emit(self, record):
        # If the record came from Rollbar's own logger don't report it
        # to Rollbar
        if record.name == rollbar.__log_name__:
            return

        level = record.levelname.lower()

        if level not in self.SUPPORTED_LEVELS:
            return

        exc_info = record.exc_info

        extra_data = {
            'args': record.args,
            'record': {
                'created': record.created,
                'funcName': record.funcName,
                'lineno': record.lineno,
                'module': record.module,
                'name': record.name,
                'pathname': record.pathname,
                'process': record.process,
                'processName': record.processName,
                'relativeCreated': record.relativeCreated,
                'thread': record.thread,
                'threadName': record.threadName
            }
        }

        extra_data.update(getattr(record, 'extra_data', {}))

        payload_data = getattr(record, 'payload_data', {})

        self._add_history(record, payload_data)

        # after we've added the history data, check to see if the
        # notify level is satisfied
        if record.levelno < self.notify_level:
            return

        # Wait until we know we're going to send a report before trying to
        # load the request
        request = getattr(record, "request", None) or rollbar.get_request()

        uuid = None
        try:
            # when not in an exception handler, exc_info == (None, None, None)
            if exc_info and exc_info[0]:
                if record.msg:
                    message_template = {
                        'body': {
                            'trace': {
                                'exception': {
                                    'description': record.getMessage()
                                }
                            }
                        }
                    }
                    payload_data = rollbar.dict_merge(
                        payload_data, message_template, silence_errors=True)

                uuid = rollbar.report_exc_info(exc_info,
                                               level=level,
                                               request=request,
                                               extra_data=extra_data,
                                               payload_data=payload_data)
            else:
                uuid = rollbar.report_message(record.getMessage(),
                                              level=level,
                                              request=request,
                                              extra_data=extra_data,
                                              payload_data=payload_data)
        except:
            self.handleError(record)
        else:
            if uuid:
                record.rollbar_uuid = uuid

    def _add_history(self, record, payload_data):
        if hasattr(self._history, 'records'):
            records = self._history.records
            history = list(records[-self.history_size:])

            if history:
                history_data = [self._build_history_data(r) for r in history]
                payload_data.setdefault('server', {})['history'] = history_data

            records.append(record)

            # prune the messages if we have too many
            self._history.records = list(records[-self.history_size:])

    def _build_history_data(self, record):
        data = {'timestamp': record.created,
                'format': record.msg,
                'args': record.args}

        if hasattr(record, 'rollbar_uuid'):
            data['uuid'] = record.rollbar_uuid

        return data



__all__ = ['add_to', 'ReporterMiddleware', 'LoggerMiddleware', 'get_current_request']

# Optional requirements:
#
# - FastAPI requires `python-multipart` package to support requests body parsing
# - `LoggerMiddleware` and `get_current_request` require `aiocontextvars` package
#   to be installed when running in Python 3.6

from .middleware import ReporterMiddleware
from .logger import LoggerMiddleware
from .routing import add_to

# Do not modify the returned request object
from rollbar.contrib.starlette import get_current_request



import sys

try:
    import fastapi

    FASTAPI_INSTALLED = True
except ImportError:
    FASTAPI_INSTALLED = False

import unittest

from rollbar.test import BaseTest

ALLOWED_PYTHON_VERSION = sys.version_info >= (3, 6)


@unittest.skipUnless(
    FASTAPI_INSTALLED and ALLOWED_PYTHON_VERSION, 'FastAPI requires Python3.6+'
)
class UtilsMiddlewareTest(BaseTest):
    def test_should_return_installed_rollbar_middlewares(self):
        from fastapi import FastAPI
        from rollbar.contrib.fastapi.utils import get_installed_middlewares
        from rollbar.contrib.fastapi import ReporterMiddleware as FastAPIMiddleware
        from rollbar.contrib.starlette import ReporterMiddleware as StarletteMiddleware
        from rollbar.contrib.asgi import ReporterMiddleware as ASGIMiddleware

        # single middleware
        app = FastAPI()
        app.add_middleware(FastAPIMiddleware)

        middlewares = get_installed_middlewares(app)

        self.assertListEqual(middlewares, [FastAPIMiddleware])

        # multiple middlewares
        app = FastAPI()
        app.add_middleware(FastAPIMiddleware)
        app.add_middleware(StarletteMiddleware)
        app.add_middleware(ASGIMiddleware)

        middlewares = get_installed_middlewares(app)

        self.assertListEqual(
            middlewares, ([ASGIMiddleware, StarletteMiddleware, FastAPIMiddleware])
        )

    def test_should_return_empty_list_if_rollbar_middlewares_not_installed(self):
        from fastapi import FastAPI
        from rollbar.contrib.fastapi.utils import get_installed_middlewares
        from rollbar.lib._async import BareMiddleware

        # no middlewares
        app = FastAPI()

        middlewares = get_installed_middlewares(app)

        self.assertListEqual(middlewares, [])

        # no Rollbar middlewares
        app = FastAPI()
        app.add_middleware(BareMiddleware)

        middlewares = get_installed_middlewares(app)

        self.assertListEqual(middlewares, [])


@unittest.skipUnless(
    FASTAPI_INSTALLED and ALLOWED_PYTHON_VERSION, 'FastAPI requires Python3.6+'
)
class UtilsBareRoutingTest(BaseTest):
    def test_should_return_true_if_has_bare_routing(self):
        from fastapi import APIRouter, FastAPI
        from rollbar.contrib.fastapi.utils import has_bare_routing

        app = FastAPI()
        self.assertTrue(has_bare_routing(app))

        router = APIRouter()
        self.assertTrue(has_bare_routing(router))

    def test_should_return_false_if_user_routes_added_to_app(self):
        from fastapi import APIRouter, FastAPI
        from rollbar.contrib.fastapi.utils import has_bare_routing

        app = FastAPI()
        self.assertTrue(has_bare_routing(app))

        @app.get('/')
        async def read_root():
            ...

        self.assertFalse(has_bare_routing(app))

    def test_should_return_false_if_user_routes_added_to_router(self):
        from fastapi import APIRouter
        from rollbar.contrib.fastapi.utils import has_bare_routing

        router = APIRouter()
        self.assertTrue(has_bare_routing(router))

        @router.get('/')
        async def read_root():
            ...

        self.assertFalse(has_bare_routing(router))

    def test_should_return_false_if_router_added_to_app(self):
        from fastapi import APIRouter, FastAPI
        from rollbar.contrib.fastapi.utils import has_bare_routing

        app = FastAPI()
        router = APIRouter()
        self.assertTrue(has_bare_routing(app))

        @router.get('/')
        async def read_root():
            ...

        app.include_router(router)

        self.assertFalse(has_bare_routing(app))

@unittest.skipUnless(
    FASTAPI_INSTALLED and ALLOWED_PYTHON_VERSION, 'FastAPI requires Python3.6+'
)
class UtilsVersionCompareTest(BaseTest):
    def test_is_current_version_higher_or_equal(self):
        # Copied from https://semver.org/#spec-item-11
        versions = [
            '1.0.0-alpha',
            '1.0.0-alpha.1',
            '1.0.0-alpha.beta',
            '1.0.0-beta',
            '1.0.0-beta.2',
            '1.0.0-beta.11',
            '1.0.0-rc.1',
            '1.0.0',
            '1.1.1',
            '1.100.0-beta2',
            '1.100.0-beta3',
        ]

        from rollbar.contrib.fastapi.utils import is_current_version_higher_or_equal

        previous_version = None
        for version in versions:
            print(f'{version} >= {previous_version}')
            if previous_version is None:
                previous_version = version
                continue
            with self.subTest(f'{version} >= {previous_version}'):
                self.assertTrue(is_current_version_higher_or_equal(version, previous_version))
            previous_version = version



"""
Tests for the RollbarHandler logging handler
"""
import copy
import json
import logging
import sys

from unittest import mock

import rollbar
from rollbar.logger import RollbarHandler

from rollbar.test import BaseTest


_test_access_token = 'aaaabbbbccccddddeeeeffff00001111'
_test_environment = 'test'
_default_settings = copy.deepcopy(rollbar.SETTINGS)

class CauseException(Exception):
    pass


class LogHandlerTest(BaseTest):
    def setUp(self):
        rollbar._initialized = False
        rollbar.SETTINGS = copy.deepcopy(_default_settings)

    def _create_logger(self):
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)

        self.rollbar_handler = RollbarHandler(_test_access_token, _test_environment)
        self.rollbar_handler.setLevel(logging.WARNING)

        logger.addHandler(self.rollbar_handler)

        return logger

    @mock.patch('rollbar.send_payload')
    def test_message_gets_formatted(self, send_payload):
        logger = self._create_logger()
        logger.warning("Hello %d %s", 1, 'world')

        payload = send_payload.call_args[0][0]

        self.assertEqual(payload['data']['body']['message']['body'], "Hello 1 world")
        self.assertEqual(payload['data']['body']['message']['args'], (1, 'world'))
        self.assertEqual(payload['data']['body']['message']['record']['name'], __name__)

    @mock.patch('rollbar.send_payload')
    def test_string_or_int_level(self, send_payload):
        logger = self._create_logger()
        logger.setLevel(logging.ERROR)
        self.rollbar_handler.setLevel('WARNING')
        logger.error("I am an error")

        payload = send_payload.call_args[0][0]

        self.assertEqual(payload['data']['level'], 'error')

        self.rollbar_handler.setLevel(logging.WARNING)
        logger.error("I am an error")

        self.assertEqual(payload['data']['level'], 'error')

    def test_request_is_get_from_log_record_if_present(self):
        logger = self._create_logger()
        # Request objects vary depending on python frameworks or packages.
        # Using a dictionary for this test is enough.
        request = {"fake": "request", "for":  "testing purporse"}

        # No need to test request parsing and payload sent,
        # just need to be sure that proper rollbar function is called
        # with passed request as argument.
        with mock.patch("rollbar.report_message") as report_message_mock:
            logger.warning("Warning message", extra={"request": request})
            self.assertEqual(report_message_mock.call_args[1]["request"], request)

        # if you call logger.exception outside of an exception
        # handler, it shouldn't try to report exc_info, since it
        # won't have any
        with mock.patch("rollbar.report_exc_info") as report_exc_info:
            with mock.patch("rollbar.report_message") as report_message_mock:
                logger.exception("Exception message", extra={"request": request})
                report_exc_info.assert_not_called()
                self.assertEqual(report_message_mock.call_args[1]["request"], request)

        with mock.patch("rollbar.report_exc_info") as report_exc_info:
            with mock.patch("rollbar.report_message") as report_message_mock:
                try:
                    raise Exception()
                except:
                    logger.exception("Exception message", extra={"request": request})
                    self.assertEqual(report_exc_info.call_args[1]["request"], request)
                    report_message_mock.assert_not_called()

    @mock.patch('rollbar.send_payload')
    def test_nested_exception_trace_chain(self, send_payload):
        logger = self._create_logger()

        def _raise_context():
            bar_local = 'bar'
            raise CauseException('bar')

        def _raise_ex():
            try:
                _raise_context()
            except CauseException as context:
                # python2 won't automatically assign this traceback...
                exc_info = sys.exc_info()
                setattr(context, '__traceback__', exc_info[2])
                try:
                    foo_local = 'foo'
                    # in python3 __context__ is automatically set when an exception is raised in an except block
                    e = Exception('foo')
                    setattr(e, '__context__', context)  # PEP-3134
                    raise e
                except:
                    logger.exception("Bad time")

        _raise_ex()

        self.assertEqual(send_payload.called, True)
        payload = send_payload.call_args[0][0]
        body = payload['data']['body']
        trace = body['trace'] if 'trace' in body else None
        trace_chain = body['trace_chain'] if 'trace_chain' in body else None
        has_only_trace_chain = trace is None and trace_chain is not None
        has_only_trace = trace is not None and trace_chain is None
        self.assertTrue(has_only_trace or has_only_trace_chain)
        if trace_chain is not None:
            self.assertEqual('Bad time', payload['data']['custom']['exception']['description'])
        if trace is not None:
            self.assertEqual('Bad time', trace['exception']['description'])

    @mock.patch('rollbar.send_payload')
    def test_not_nested_exception_trace_chain(self, send_payload):
        logger = self._create_logger()

        def _raise_context():
            bar_local = 'bar'
            raise CauseException('bar')

        def _raise_ex():
            try:
                _raise_context()
            except:
                logger.exception("Bad time")

        _raise_ex()

        self.assertEqual(send_payload.called, True)
        payload = send_payload.call_args[0][0]
        body = payload['data']['body']
        trace = body['trace'] if 'trace' in body else None
        trace_chain = body['trace_chain'] if 'trace_chain' in body else None
        has_only_trace_chain = trace is None and trace_chain is not None
        has_only_trace = trace is not None and trace_chain is None
        self.assertTrue(has_only_trace or has_only_trace_chain)
        if trace_chain is not None:
            self.assertEqual('Bad time', payload['data']['custom']['exception']['description'])
        if trace is not None:
            self.assertEqual('Bad time', trace['exception']['description'])



__all__ = ['LoggerMiddleware']

import logging
import sys

from starlette import __version__
from starlette.types import ASGIApp, Receive, Scope, Send

from rollbar.contrib.asgi import ReporterMiddleware as ASGIReporterMiddleware
from rollbar.contrib.asgi.integration import integrate
from rollbar.contrib.starlette.requests import store_current_request

log = logging.getLogger(__name__)


@integrate(framework_name=f'starlette {__version__}')
class LoggerMiddleware(ASGIReporterMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        if sys.version_info < (3, 6):
            log.error(
                'LoggerMiddleware requires Python 3.7+ (or 3.6 with `aiocontextvars` package)'
            )
            raise RuntimeError(
                'LoggerMiddleware requires Python 3.7+ (or 3.6 with `aiocontextvars` package)'
            )

        super().__init__(app)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        store_current_request(scope, receive)

        await self.app(scope, receive, send)



"""
django-rollbar context processor

To install, add the following in your settings.py:
1. add 'rollbar.contrib.django.context_processors.rollbar_settings' to TEMPLATE_CONTEXT_PROCESSORS
2. add a section like this:
ROLLBAR = {
    'client_access_token': 'tokengoeshere',
}
3. you can now access your rollbar settings as rollbar_settings from within your django templates

See README.rst for full installation and configuration instructions.
"""

from django.conf import settings


def rollbar_settings(request):
    """Grabs the rollbar settings to make them available to templates."""
    if not hasattr(settings, 'ROLLBAR'):
        return {}
    return {'rollbar_settings': settings.ROLLBAR}



import optparse
import sys

import rollbar

VERSION = '0.1'

verbose = False

def _gen_report_message(level):
    def _wrapped(lines):
        line_data = '\n'.join(lines)
        if verbose:
            print('Rollbar [%s]: %s' % (level, line_data))
        rollbar.report_message(line_data, level=level, extra_data={'cli_version': VERSION})
    return _wrapped


CMDS = {
    'debug': _gen_report_message('debug'),
    'info': _gen_report_message('info'),
    'warning': _gen_report_message('warning'),
    'error': _gen_report_message('error'),
    'critical': _gen_report_message('critical'),
}

def main():
    global verbose
    parser = optparse.OptionParser(version='%prog version ' + VERSION)
    parser.add_option('-t', '--access_token',
                      dest='access_token',
                      help="You project's access token from rollbar.com.",
                      metavar='ACCESS_TOKEN')
    parser.add_option('-e', '--environment',
                      dest='environment',
                      help="The environment to report errors and messages to. \
                      Can be any string; suggestions: 'production', 'development', 'staging'",
                      metavar='ENVIRONMENT')
    parser.add_option('-u', '--url',
                      dest='endpoint_url',
                      help="The Rollbar API endpoint url to send data to.",
                      metavar='ENDPOINT_URL',
                      default=rollbar.DEFAULT_ENDPOINT)
    parser.add_option('-m', '--handler',
                      dest='handler',
                      help="The method in which to report errors.",
                      metavar='HANDLER',
                      choices=["thread", "blocking", "agent"],
                      default="blocking")
    parser.add_option('-v', '--verbose',
                      dest='verbose',
                      help="Print verbose output.",
                      action='store_true',
                      default=False)

    options, args = parser.parse_args()

    access_token = options.access_token
    env = options.environment
    endpoint = options.endpoint_url
    handler = options.handler
    verbose = options.verbose

    if not access_token:
        parser.error('missing access_token')
    if not env:
        parser.error('missing environment')

    rollbar.init(access_token, environment=env, endpoint=endpoint, handler=handler)

    def _do_cmd(cmd_name, line):
        cmd = CMDS.get(cmd_name.lower())
        if cmd:
            cmd([line])
            return True

        return False

    if len(args) > 1:
        sent = _do_cmd(args[0], ' '.join(args[1:]))
        sys.exit(0 if sent else 1)

    cur_cmd_name = None
    try:
        cur_line = sys.stdin.readline()
        while cur_line:
            cur_line = cur_line.strip()
            parts = cur_line.split(' ')

            if parts:
                cur_cmd_name = parts[0]
                parts = parts[1:]
                _do_cmd(cur_cmd_name, ' '.join(parts))

            cur_line = sys.stdin.readline()
    except (KeyboardInterrupt, SystemExit) as e:
        pass



import base64
import copy
import json
import socket
import threading
import uuid

import sys

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from unittest import mock

import unittest

import rollbar
from rollbar.lib import string_types

from rollbar.test import BaseTest

try:
    eval("""
        def _anonymous_tuple_func(x, (a, b), y):
            ret = x + a + b + y
            breakme()
            return ret
    """)
except SyntaxError:
    _anonymous_tuple_func = None


_test_access_token = 'aaaabbbbccccddddeeeeffff00001111'
_default_settings = copy.deepcopy(rollbar.SETTINGS)


class RollbarTest(BaseTest):
    def setUp(self):
        rollbar._initialized = False
        rollbar.SETTINGS = copy.deepcopy(_default_settings)
        rollbar.init(_test_access_token, locals={'enabled': True}, dummy_key='asdf', handler='blocking', timeout=12345)

    def test_merged_settings(self):
        expected = {'enabled': True, 'sizes': rollbar.DEFAULT_LOCALS_SIZES, 'safe_repr': True, 'scrub_varargs': True, 'safelisted_types': [], 'whitelisted_types': []}
        self.assertDictEqual(rollbar.SETTINGS['locals'], expected)
        self.assertEqual(rollbar.SETTINGS['timeout'], 12345)
        self.assertEqual(rollbar.SETTINGS['dummy_key'], 'asdf')

    def test_default_configuration(self):
        self.assertEqual(rollbar.SETTINGS['access_token'], _test_access_token)
        self.assertEqual(rollbar.SETTINGS['environment'], 'production')

    @mock.patch('rollbar.send_payload')
    def test_disabled(self, send_payload):
        rollbar.SETTINGS['enabled'] = False

        rollbar.report_message('foo')
        try:
            raise Exception('foo')
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, False)

    def test_server_data(self):
        server_data = rollbar._build_server_data()

        self.assertIn('host', server_data)
        self.assertIn('argv', server_data)
        self.assertNotIn('branch', server_data)
        self.assertNotIn('root', server_data)

        rollbar.SETTINGS['branch'] = 'master'
        rollbar.SETTINGS['root'] = '/home/test/'

        server_data = rollbar._build_server_data()

        self.assertIn('host', server_data)
        self.assertIn('argv', server_data)
        self.assertEqual(server_data['branch'], 'master')
        self.assertEqual(server_data['root'], '/home/test/')

    def test_wsgi_request_data(self):
        request = {
            'CONTENT_LENGTH': str(len('body body body')),
            'CONTENT_TYPE': '',
            'DOCUMENT_URI': '/api/test',
            'GATEWAY_INTERFACE': 'CGI/1.1',
            'HTTP_CONNECTION': 'close',
            'HTTP_HOST': 'example.com',
            'HTTP_USER_AGENT': 'Agent',
            'PATH_INFO': '/api/test',
            'QUERY_STRING': 'format=json&param1=value1&param2=value2',
            'REMOTE_ADDR': '127.0.0.1',
            'REQUEST_METHOD': 'GET',
            'SCRIPT_NAME': '',
            'SERVER_ADDR': '127.0.0.1',
            'SERVER_NAME': 'example.com',
            'SERVER_PORT': '80',
            'SERVER_PROTOCOL': 'HTTP/1.1',
            'wsgi.input': StringIO('body body body'),
            'wsgi.multiprocess': True,
            'wsgi.multithread': False,
            'wsgi.run_once': False,
            'wsgi.url_scheme': 'http',
            'wsgi.version': (1, 0)
        }
        data = rollbar._build_wsgi_request_data(request)
        self.assertEqual(data['url'], 'http://example.com/api/test?format=json&param1=value1&param2=value2')
        self.assertEqual(data['user_ip'], '127.0.0.1')
        self.assertEqual(data['method'], 'GET')
        self.assertEqual(data['body'], 'body body body')
        self.assertDictEqual(data['GET'], {'format': 'json', 'param1': 'value1', 'param2': 'value2'})
        self.assertDictEqual(data['headers'], {'Connection': 'close', 'Host': 'example.com', 'User-Agent': 'Agent'})

    def test_starlette_request_data(self):
        try:
            from starlette.requests import Request
        except ImportError:
            self.skipTest('Requires Starlette to be installed')

        scope = {
            'type': 'http',
            'client': ('127.0.0.1', 1453),
            'headers': [
                (b'accept', b'*/*'),
                (b'content-type', b'application/x-www-form-urlencoded'),
                (b'host', b'example.com'),
                (b'user-agent', b'Agent'),
            ],
            'http_version': '1.1',
            'method': 'GET',
            'path': '/api/test',
            'path_params': {'param': 'test'},
            'query_params': {
                'format': 'json',
                'param1': 'value1',
                'param2': 'value2',
            },
            'query_string': b'format=json&param1=value1&param2=value2',
            'scheme': 'http',
            'server': ('example.com', 80),
            'url': {'path': 'example.com'},
        }
        request = Request(scope)
        data = rollbar._build_starlette_request_data(request)

        self.assertEqual(data['url'], 'http://example.com/api/test?format=json&param1=value1&param2=value2')
        self.assertEqual(data['user_ip'], '127.0.0.1')
        self.assertEqual(data['method'], 'GET')
        self.assertDictEqual(data['params'], {'param': 'test'})
        self.assertDictEqual(data['GET'], {'format': 'json', 'param1': 'value1', 'param2': 'value2'})
        self.assertDictEqual(
            data['headers'],
            {
                'accept': '*/*',
                'content-type': 'application/x-www-form-urlencoded',
                'host': 'example.com',
                'user-agent': 'Agent',
            },
        )

    def test_starlette_request_data_with_consumed_body(self):
        try:
            from starlette.requests import Request
        except ImportError:
            self.skipTest('Requires Starlette to be installed')
        from rollbar.lib._async import async_receive, run

        rollbar.SETTINGS['include_request_body'] = True
        body = b'body body body'
        scope = {
            'type': 'http',
            'headers': [
                (b'content-type', b'text/html'),
                (b'content-length', str(len(body)).encode('latin-1')),
            ],
            'method': 'GET',
            'path': '/api/test',
            'query_string': b'',
        }
        receive = async_receive(
            {'type': 'http.request', 'body': body, 'mode_body': False}
        )
        request = Request(scope, receive)

        # Consuming body in Starlette middleware is currently disabled
        run(request.body()) # await request.body()

        data = rollbar._build_starlette_request_data(request)

        self.assertEqual(data['body'], body.decode('latin-1'))

    def test_starlette_request_data_empty_values(self):
        try:
            from starlette.requests import Request
        except ImportError:
            self.skipTest('Requires Starlette to be installed')

        scope = {
            'type': 'http',
            'client': ('127.0.0.1', 1453),
            'headers': [
                (b'content-type', b'text/html'),
            ],
            'method': 'GET',
            'query_string': b'',
            'path': '',
        }
        request = Request(scope)

        data = rollbar._build_starlette_request_data(request)

        self.assertFalse('GET' in data)
        self.assertFalse('url' in data)
        self.assertFalse('params' in data)
        self.assertTrue('headers' in data)
        self.assertEqual(data['user_ip'], scope['client'][0])
        self.assertEqual(data['method'], scope['method'])

    def test_fastapi_request_data(self):
        try:
            from fastapi.requests import Request
        except ImportError:
            self.skipTest('Requires FastAPI to be installed')

        scope = {
            'type': 'http',
            'client': ('127.0.0.1', 1453),
            'headers': [
                (b'accept', b'*/*'),
                (b'content-type', b'application/x-www-form-urlencoded'),
                (b'host', b'example.com'),
                (b'user-agent', b'Agent'),
            ],
            'http_version': '1.1',
            'method': 'GET',
            'path': '/api/test',
            'path_params': {'param': 'test'},
            'query_params': {
                'format': 'json',
                'param1': 'value1',
                'param2': 'value2',
            },
            'query_string': b'format=json&param1=value1&param2=value2',
            'scheme': 'http',
            'server': ('example.com', 80),
            'url': {'path': 'example.com'},
        }
        request = Request(scope)
        data = rollbar._build_fastapi_request_data(request)

        self.assertEqual(data['url'], 'http://example.com/api/test?format=json&param1=value1&param2=value2')
        self.assertEqual(data['user_ip'], '127.0.0.1')
        self.assertEqual(data['method'], 'GET')
        self.assertDictEqual(data['params'], {'param': 'test'})
        self.assertDictEqual(data['GET'], {'format': 'json', 'param1': 'value1', 'param2': 'value2'})
        self.assertDictEqual(
            data['headers'],
            {
                'accept': '*/*',
                'content-type': 'application/x-www-form-urlencoded',
                'host': 'example.com',
                'user-agent': 'Agent',
            },
        )

    def test_fastapi_request_data_with_consumed_body(self):
        try:
            from fastapi import Request
        except ImportError:
            self.skipTest('Requires FastAPI to be installed')
        from rollbar.lib._async import async_receive, run

        rollbar.SETTINGS['include_request_body'] = True
        body = b'body body body'
        scope = {
            'type': 'http',
            'headers': [
                (b'content-type', b'text/html'),
                (b'content-length', str(len(body)).encode('latin-1')),
            ],
            'method': 'GET',
            'path': '/api/test',
            'query_string': b'',
        }
        receive = async_receive(
            {'type': 'http.request', 'body': body, 'mode_body': False}
        )
        request = Request(scope, receive)

        # Consuming body in FastAPI middlewares is currently disabled
        run(request.body()) # await request.body()

        data = rollbar._build_fastapi_request_data(request)

        self.assertEqual(data['body'], body.decode('latin-1'))

    def test_fastapi_request_data_empty_values(self):
        try:
            from fastapi import Request
        except ImportError:
            self.skipTest('Requires FastAPI to be installed')

        scope = {
            'type': 'http',
            'client': ('127.0.0.1', 1453),
            'headers': [
                (b'content-type', b'text/html'),
            ],
            'method': 'GET',
            'query_string': b'',
            'path': '',
        }
        request = Request(scope)

        data = rollbar._build_fastapi_request_data(request)

        self.assertFalse('GET' in data)
        self.assertFalse('url' in data)
        self.assertFalse('params' in data)
        self.assertTrue('headers' in data)
        self.assertEqual(data['user_ip'], scope['client'][0])
        self.assertEqual(data['method'], scope['method'])

    def test_django_build_person_data(self):
        try:
            import django
            from django.conf import settings
        except ImportError:
            self.skipTest('Requires Django to be installed')
        else:
            settings.configure(
                INSTALLED_APPS=['django.contrib.auth', 'django.contrib.contenttypes']
            )
            if django.VERSION >= (1, 7):
                django.setup()

        from django.contrib.auth.models import User
        from django.http.request import HttpRequest

        request = HttpRequest()
        request.user = User()
        request.user.id = 123
        request.user.username = 'admin'
        request.user.email = 'admin@example.org'

        data = rollbar._build_person_data(request)

        self.assertDictEqual(
            data, {'id': '123', 'username': 'admin', 'email': 'admin@example.org'}
        )

    def test_starlette_build_person_data_if_user_authenticated(self):
        try:
            from starlette.authentication import SimpleUser
            from starlette.requests import Request
        except ImportError:
            self.skipTest('Requires Starlette to be installed')

        # Implement interface with the id attribute
        class User(SimpleUser):
            counter = 0

            def __init__(self, username, email):
                super().__init__(username)
                self.email = email

                User.counter += 1
                self.id = User.counter

        scope = {'type': 'http'}
        request = Request(scope)
        # Make the user authenticated
        request.scope['user'] = User('admin', 'admin@example.org')

        data = rollbar._build_person_data(request)

        self.assertDictEqual(
            data, {'id': '1', 'username': 'admin', 'email': 'admin@example.org'}
        )

    def test_starlette_failsafe_build_person_data_if_user_not_authenticated(self):
        try:
            from starlette.requests import Request
        except ImportError:
            self.skipTest('Requires Starlette to be installed')

        scope = {'type': 'http'}
        request = Request(scope)

        data = rollbar._build_person_data(request)

        self.assertIsNone(data)

    @unittest.skipUnless(sys.version_info >= (3, 6), 'Python3.6+ required')
    def test_get_request_starlette_middleware(self):
        try:
            from starlette.applications import Starlette
            from starlette.middleware import Middleware
            from starlette.responses import PlainTextResponse
            from starlette.routing import Route
            from starlette.testclient import TestClient
        except ImportError:
            self.skipTest('Requires Starlette package')
        from rollbar.contrib.starlette import ReporterMiddleware

        def root(starlette_request):
            current_request = rollbar.get_request()

            self.assertEqual(current_request, starlette_request)

            return PlainTextResponse("bye bye")

        routes = [Route('/{param}', root)]
        middleware = [Middleware(ReporterMiddleware)]
        app = Starlette(routes=routes, middleware=middleware)
        client = TestClient(app)
        response = client.get('/test?param1=value1&param2=value2')

        self.assertEqual(response.status_code, 200)

    @unittest.skipUnless(sys.version_info >= (3, 6), 'Python3.6+ required')
    def test_get_request_starlette_logger(self):
        try:
            from starlette.applications import Starlette
            from starlette.middleware import Middleware
            from starlette.responses import PlainTextResponse
            from starlette.routing import Route
            from starlette.testclient import TestClient
        except ImportError:
            self.skipTest('Requires Starlette package')
        from rollbar.contrib.starlette import ReporterMiddleware

        def root(starlette_request):
            current_request = rollbar.get_request()

            self.assertEqual(current_request, starlette_request)

            return PlainTextResponse("bye bye")

        routes = [Route('/{param}', root)]
        middleware = [Middleware(ReporterMiddleware)]
        app = Starlette(routes=routes, middleware=middleware)
        client = TestClient(app)
        response = client.get('/test?param1=value1&param2=value2')

        self.assertEqual(response.status_code, 200)

    @unittest.skipUnless(sys.version_info >= (3, 6), 'Python3.6+ required')
    def test_get_request_fastapi_middleware(self):
        try:
            from fastapi import FastAPI, Request
            from fastapi.testclient import TestClient
        except ImportError:
            self.skipTest('Requires FastaAPI package')
        from rollbar.contrib.fastapi import ReporterMiddleware

        app = FastAPI()
        app.add_middleware(ReporterMiddleware)

        @app.get('/{param}')
        def root(param, fastapi_request: Request):
            current_request = rollbar.get_request()

            self.assertEqual(current_request, fastapi_request)

        root = fastapi_add_route_with_request_param(
            app, root, '/{param}', 'fastapi_request'
        )

        client = TestClient(app)
        response = client.get('/test?param1=value1&param2=value2')

        self.assertEqual(response.status_code, 200)

    @unittest.skipUnless(sys.version_info >= (3, 6), 'Python3.6+ required')
    def test_get_request_fastapi_logger(self):
        try:
            from fastapi import FastAPI, Request
            from fastapi.testclient import TestClient
        except ImportError:
            self.skipTest('Requires FastaAPI package')
        from rollbar.contrib.fastapi import ReporterMiddleware

        app = FastAPI()
        app.add_middleware(ReporterMiddleware)

        @app.get('/{param}')
        def root(fastapi_request: Request):
            current_request = rollbar.get_request()

            self.assertEqual(current_request, fastapi_request)

        root = fastapi_add_route_with_request_param(
            app, root, '/{param}', 'fastapi_request'
        )

        client = TestClient(app)
        response = client.get('/test?param1=value1&param2=value2')

        self.assertEqual(response.status_code, 200)

    @unittest.skipUnless(sys.version_info >= (3, 6), 'Python3.6+ required')
    def test_get_request_fastapi_router(self):
        try:
            import fastapi
            from fastapi import FastAPI, Request
            from fastapi.testclient import TestClient
        except ImportError:
            self.skipTest('Requires FastAPI package')
        from rollbar.contrib.fastapi import add_to as rollbar_add_to

        if fastapi.__version__ < '0.41.0':
            self.skipTest('Requires FastAPI 0.41.0+')

        app = FastAPI()
        rollbar_add_to(app)

        @app.get('/{param}')
        def root(fastapi_request: Request):
            current_request = rollbar.get_request()

            self.assertEqual(current_request, fastapi_request)

        root = fastapi_add_route_with_request_param(
            app, root, '/{param}', 'fastapi_request'
        )

        client = TestClient(app)
        response = client.get('/test?param1=value1&param2=value2')

        self.assertEqual(response.status_code, 200)

    @mock.patch('rollbar.send_payload')
    def test_report_exception(self, send_payload):

        def _raise():
            try:
                raise Exception('foo')
            except:
                rollbar.report_exc_info()

        _raise()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertEqual(payload['access_token'], _test_access_token)
        self.assertIn('body', payload['data'])
        self.assertIn('trace', payload['data']['body'])
        self.assertNotIn('trace_chain', payload['data']['body'])
        self.assertIn('exception', payload['data']['body']['trace'])
        self.assertEqual(payload['data']['body']['trace']['exception']['message'], 'foo')
        self.assertEqual(payload['data']['body']['trace']['exception']['class'], 'Exception')

        self.assertNotIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('locals', payload['data']['body']['trace']['frames'][-1])

    @mock.patch('rollbar._post_api')
    def test_lambda_function_good(self, _post_api):
        rollbar.SETTINGS['handler'] = 'thread'
        fake_event = {'a': 42}
        fake_context = MockLambdaContext(99)
        @rollbar.lambda_function
        def my_lambda_func(event, context):
            return [event['a'], context.x]

        result = my_lambda_func(fake_event, fake_context)

        self.assertEqual(len(result), 2)
        self.assertEqual(result[0], 42)
        self.assertEqual(result[1], 99)
        self.assertEqual(_post_api.called, False)

        rollbar._CURRENT_LAMBDA_CONTEXT = None
        rollbar.SETTINGS['handler'] = 'blocking'

    @mock.patch('rollbar._post_api')
    def test_lambda_function_bad(self, _post_api):
        rollbar.SETTINGS['handler'] = 'thread'
        fake_event = {'a': 42}
        fake_context = MockLambdaContext(99)
        @rollbar.lambda_function
        def my_lambda_func(event, context):
            raise event['a']

        result = None
        try:
            result = my_lambda_func(fake_event, fake_context)
        except:
            pass

        self.assertEqual(result, None)
        self.assertEqual(_post_api.called, True)

        rollbar._CURRENT_LAMBDA_CONTEXT = None
        rollbar.SETTINGS['handler'] = 'blocking'

    @mock.patch('rollbar._post_api')
    def test_lambda_function_method_good(self, _post_api):
        rollbar.SETTINGS['handler'] = 'thread'
        fake_event = {'a': 42}
        fake_context = MockLambdaContext(99)

        class LambdaClass(object):
            def __init__(self):
                self.a = 13

            def my_lambda_func(self, event, context):
                return [event['a'], context.x, self.a]

        app = LambdaClass()
        app.my_lambda_func = rollbar.lambda_function(app.my_lambda_func)
        result = app.my_lambda_func(fake_event, fake_context)

        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], 42)
        self.assertEqual(result[1], 99)
        self.assertEqual(result[2], 13)
        self.assertEqual(_post_api.called, False)

        rollbar._CURRENT_LAMBDA_CONTEXT = None
        rollbar.SETTINGS['handler'] = 'blocking'

    @mock.patch('rollbar._post_api')
    def test_lambda_function_method_bad(self, _post_api):
        rollbar.SETTINGS['handler'] = 'thread'
        fake_event = {'a': 42}
        fake_context = MockLambdaContext(99)

        class LambdaClass(object):
            def __init__(self):
                self.a = 13

            def my_lambda_func(self, event, context):
                raise self.a

        app = LambdaClass()
        app.my_lambda_func = rollbar.lambda_function(app.my_lambda_func)

        result = None
        try:
            result = app.my_lambda_func(fake_event, fake_context)
        except:
            pass

        self.assertEqual(result, None)
        self.assertEqual(_post_api.called, True)

        rollbar._CURRENT_LAMBDA_CONTEXT = None
        rollbar.SETTINGS['handler'] = 'blocking'

    @mock.patch('rollbar.send_payload')
    def test_report_exception_with_cause(self, send_payload):
        def _raise_cause():
            bar_local = 'bar'
            raise CauseException('bar')

        def _raise_ex():
            try:
                _raise_cause()
            except CauseException as cause:
                # python2 won't automatically assign this traceback...
                exc_info = sys.exc_info()
                setattr(cause, '__traceback__', exc_info[2])

                try:
                    foo_local = 'foo'
                    # in python3 this would normally be expressed as
                    # raise Exception('foo') from cause
                    e = Exception('foo')
                    setattr(e, '__cause__', cause)  # PEP-3134
                    raise e
                except:
                    rollbar.report_exc_info()

        _raise_ex()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertEqual(payload['access_token'], _test_access_token)
        self.assertIn('body', payload['data'])
        self.assertNotIn('trace', payload['data']['body'])
        self.assertIn('trace_chain', payload['data']['body'])
        self.assertEqual(2, len(payload['data']['body']['trace_chain']))

        self.assertIn('exception', payload['data']['body']['trace_chain'][0])
        self.assertEqual(payload['data']['body']['trace_chain'][0]['exception']['message'], 'foo')
        self.assertEqual(payload['data']['body']['trace_chain'][0]['exception']['class'], 'Exception')
        self.assertEqual(payload['data']['body']['trace_chain'][0]['frames'][-1]['locals']['foo_local'], 'foo')

        self.assertIn('exception', payload['data']['body']['trace_chain'][1])
        self.assertEqual(payload['data']['body']['trace_chain'][1]['exception']['message'], 'bar')
        self.assertEqual(payload['data']['body']['trace_chain'][1]['exception']['class'], 'CauseException')
        self.assertEqual(payload['data']['body']['trace_chain'][1]['frames'][-1]['locals']['bar_local'], 'bar')

    @mock.patch('rollbar.send_payload')
    def test_report_exception_with_same_exception_as_cause(self, send_payload):
        cause_exc = CauseException('bar')

        def _raise_cause():
            bar_local = 'bar'
            raise cause_exc

        def _raise_ex():
            try:
                _raise_cause()
            except CauseException as cause:
                # python2 won't automatically assign this traceback...
                exc_info = sys.exc_info()
                setattr(cause, '__traceback__', exc_info[2])

                try:
                    foo_local = 'foo'
                    # in python3 this would normally be expressed as
                    # raise cause from cause
                    setattr(cause, '__cause__', cause)  # PEP-3134
                    raise cause
                except:
                    rollbar.report_exc_info()

        ex_raiser = threading.Thread(target=_raise_ex)
        ex_raiser.daemon = True
        ex_raiser.start()
        # 0.5 seconds ought be enough for any modern computer to get into the
        # cyclical parts of the code, but not so long as to collect a lot of
        # objects in memory
        ex_raiser.join(timeout=0.5)

        if ex_raiser.is_alive():
            # This breaks the circular reference, allowing thread to exit and
            # to be joined
            cause_exc.__cause__ = None
            ex_raiser.join()
            self.fail('Cyclic reference in rollbar._walk_trace_chain()')

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertEqual(payload['access_token'], _test_access_token)
        self.assertIn('body', payload['data'])
        self.assertNotIn('trace', payload['data']['body'])
        self.assertIn('trace_chain', payload['data']['body'])
        self.assertEqual(2, len(payload['data']['body']['trace_chain']))

        self.assertIn('exception', payload['data']['body']['trace_chain'][0])
        self.assertEqual(payload['data']['body']['trace_chain'][0]['exception']['message'], 'bar')
        self.assertEqual(payload['data']['body']['trace_chain'][0]['exception']['class'], 'CauseException')
        frames = payload['data']['body']['trace_chain'][0]['frames']
        self.assertEqual(payload['data']['body']['trace_chain'][0]['frames'][0]['locals']['foo_local'], 'foo')

        self.assertIn('exception', payload['data']['body']['trace_chain'][1])
        self.assertEqual(payload['data']['body']['trace_chain'][1]['exception']['message'], 'bar')
        self.assertEqual(payload['data']['body']['trace_chain'][1]['exception']['class'], 'CauseException')
        self.assertEqual(payload['data']['body']['trace_chain'][1]['frames'][-1]['locals']['bar_local'], 'bar')

    @mock.patch('rollbar.send_payload')
    def test_report_exception_with_context(self, send_payload):
        def _raise_context():
            bar_local = 'bar'
            raise CauseException('bar')

        def _raise_ex():
            try:
                _raise_context()
            except CauseException as context:
                # python2 won't automatically assign this traceback...
                exc_info = sys.exc_info()
                setattr(context, '__traceback__', exc_info[2])

                try:
                    foo_local = 'foo'
                    # in python3 __context__ is automatically set when an exception is raised in an except block
                    e = Exception('foo')
                    setattr(e, '__context__', context)  # PEP-3134
                    raise e
                except:
                    rollbar.report_exc_info()

        _raise_ex()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertEqual(payload['access_token'], _test_access_token)
        self.assertIn('body', payload['data'])
        self.assertNotIn('trace', payload['data']['body'])
        self.assertIn('trace_chain', payload['data']['body'])
        self.assertEqual(2, len(payload['data']['body']['trace_chain']))

        self.assertIn('exception', payload['data']['body']['trace_chain'][0])
        self.assertEqual(payload['data']['body']['trace_chain'][0]['exception']['message'], 'foo')
        self.assertEqual(payload['data']['body']['trace_chain'][0]['exception']['class'], 'Exception')
        self.assertEqual(payload['data']['body']['trace_chain'][0]['frames'][-1]['locals']['foo_local'], 'foo')

        self.assertIn('exception', payload['data']['body']['trace_chain'][1])
        self.assertEqual(payload['data']['body']['trace_chain'][1]['exception']['message'], 'bar')
        self.assertEqual(payload['data']['body']['trace_chain'][1]['exception']['class'], 'CauseException')
        self.assertEqual(payload['data']['body']['trace_chain'][1]['frames'][-1]['locals']['bar_local'], 'bar')

    @mock.patch('rollbar.send_payload')
    def test_exception_filters(self, send_payload):

        rollbar.SETTINGS['exception_level_filters'] = [
            (OSError, 'ignored'),
            ('rollbar.ApiException', 'ignored'),
            ('bogus.DoesntExist', 'ignored'),
        ]

        def _raise_exception():
            try:
                raise Exception('foo')
            except:
                rollbar.report_exc_info()

        def _raise_os_error():
            try:
                raise OSError('bar')
            except:
                rollbar.report_exc_info()

        def _raise_api_exception():
            try:
                raise rollbar.ApiException('bar')
            except:
                rollbar.report_exc_info()

        _raise_exception()
        self.assertTrue(send_payload.called)

        _raise_os_error()
        self.assertEqual(1, send_payload.call_count)

        _raise_api_exception()
        self.assertEqual(1, send_payload.call_count)

    @mock.patch('rollbar.send_payload')
    def test_report_messsage(self, send_payload):
        rollbar.report_message('foo')

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertEqual(payload['access_token'], _test_access_token)
        self.assertIn('body', payload['data'])
        self.assertIn('message', payload['data']['body'])
        self.assertIn('body', payload['data']['body']['message'])
        self.assertEqual(payload['data']['body']['message']['body'], 'foo')

    @mock.patch('rollbar.send_payload')
    def test_uuid(self, send_payload):
        uuid = rollbar.report_message('foo')

        payload = send_payload.call_args[0][0]

        self.assertEqual(payload['data']['uuid'], uuid)

    @mock.patch('rollbar.send_payload')
    def test_report_exc_info_level(self, send_payload):

        try:
            raise Exception('level_error')
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)
        payload = send_payload.call_args[0][0]
        self.assertEqual(payload['data']['level'], 'error')

        try:
            raise Exception('level_info')
        except:
            rollbar.report_exc_info(level='info')

        self.assertEqual(send_payload.called, True)
        payload = send_payload.call_args[0][0]
        self.assertEqual(payload['data']['level'], 'info')

        # payload takes precendence over 'level'
        try:
            raise Exception('payload_warn')
        except:
            rollbar.report_exc_info(level='info', payload_data={'level': 'warn'})

        self.assertEqual(send_payload.called, True)
        payload = send_payload.call_args[0][0]
        self.assertEqual(payload['data']['level'], 'warn')

    @mock.patch('rollbar.send_payload')
    def test_report_exc_info_nones(self, send_payload):

        rollbar.report_exc_info(exc_info=(None, None, None))

        self.assertEqual(send_payload.called, True)
        payload = send_payload.call_args[0][0]
        self.assertEqual(payload['data']['level'], 'error')

    @mock.patch('rollbar._send_failsafe')
    @mock.patch('rollbar.lib.transport.post',
                side_effect=lambda *args, **kw: MockResponse({'status': 'Payload Too Large'}, 413))
    def test_trigger_failsafe(self, post, _send_failsafe):
        rollbar.report_message('derp')
        self.assertEqual(_send_failsafe.call_count, 1)

        try:
            raise Exception('trigger_failsafe')
        except:
            rollbar.report_exc_info()
            self.assertEqual(_send_failsafe.call_count, 2)

    @mock.patch('rollbar._send_failsafe')
    @mock.patch('rollbar.lib.transport.post',
                side_effect=lambda *args, **kw: MockRawResponse('<html>\r\n' \
                                                             '<head><title>502 Bad Gateway</title></head>\r\n' \
                                                             '<body bgcolor="white">\r\n' \
                                                             '<center><h1>502 Bad Gateway</h1></center>\r\n' \
                                                             '<hr><center>nginx</center>\r\n' \
                                                             '</body>\r\n' \
                                                             '</html>\r\n', 502))
    def test_502_failsafe(self, post, _send_failsafe):
        rollbar.report_message('derp')
        # self.assertEqual(_send_failsafe.call_count, 1)

        try:
            raise Exception('trigger_failsafe')
        except:
            rollbar._post_api('/api/1/item', {'derp'})

    @mock.patch('rollbar.send_payload')
    def test_send_failsafe(self, send_payload):
        test_uuid = str(uuid.uuid4())
        test_host = socket.gethostname()
        test_data = {
            'access_token': _test_access_token,
            'data': {
                'body': {
                    'message': {
                        'body': 'Failsafe from pyrollbar: test message. '
                                'Original payload may be found in your server '
                                'logs by searching for the UUID.'
                    }
                },
                'failsafe': True,
                'level': 'error',
                'custom': {
                    'orig_host': test_host,
                    'orig_uuid': test_uuid
                },
                'environment': rollbar.SETTINGS['environment'],
                'internal': True,
                'notifier': rollbar.SETTINGS['notifier']
            }
        }

        rollbar._send_failsafe('test message', test_uuid, test_host)
        self.assertEqual(send_payload.call_count, 1)
        self.assertEqual(send_payload.call_args[0][0], test_data)

    @mock.patch('rollbar.log.exception')
    @mock.patch('rollbar.send_payload', side_effect=Exception('Monkey Business!'))
    def test_fail_to_send_failsafe(self, send_payload, mock_log):
        test_uuid = str(uuid.uuid4())
        test_host = socket.gethostname()
        rollbar._send_failsafe('test message', test_uuid, test_host)
        self.assertEqual(mock_log.call_count, 1)

    @unittest.skipUnless(rollbar.AsyncHTTPClient, 'Requires async handler to be installed')
    @mock.patch('rollbar._send_payload_async')
    def test_async_handler(self, send_payload_async):
        def _raise():
            try:
                raise Exception('foo')
            except:
                rollbar.report_exc_info()

        rollbar.SETTINGS['handler'] = 'async'
        _raise()

        send_payload_async.assert_called_once()

    @unittest.skipUnless(rollbar.httpx, 'Requires HTTPX to be installed')
    @mock.patch('rollbar._send_payload_httpx')
    def test_httpx_handler(self, send_payload_httpx):
        def _raise():
            try:
                raise Exception('foo')
            except:
                rollbar.report_exc_info()

        rollbar.SETTINGS['handler'] = 'async'
        _raise()

        send_payload_httpx.assert_called_once()

    @unittest.skipUnless(sys.version_info >= (3, 6), 'assert_called_once support requires Python3.6+')
    @mock.patch('rollbar._send_payload_thread_pool')
    def test_thread_pool_handler(self, send_payload_thread_pool):
        def _raise():
            try:
                raise Exception('foo')
            except:
                rollbar.report_exc_info()
        rollbar.SETTINGS['handler'] = 'thread_pool'
        _raise()

        send_payload_thread_pool.assert_called_once()

    @unittest.skipUnless(sys.version_info >= (3, 2), 'concurrent.futures support requires Python3.2+')
    def test_thread_pool_submit(self):
        from rollbar.lib.thread_pool import init_pool, submit
        init_pool(1)
        ran = {'nope': True}  # dict used so it is not shadowed in run

        def run(payload_str, access_token):
            ran['nope'] = False

        submit(run, 'foo', 'bar')
        self.assertFalse(ran['nope'])


    @mock.patch('rollbar.send_payload')
    def test_args_constructor(self, send_payload):

        class tmp(object):
            def __init__(self, arg1):
                self.arg1 = arg1
                foo()

        try:
            t = tmp(33)
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        self.assertEqual('arg1', payload['data']['body']['trace']['frames'][-1]['argspec'][1])
        self.assertEqual(33, payload['data']['body']['trace']['frames'][-1]['locals']['arg1'])

    @mock.patch('rollbar.send_payload')
    def test_failed_locals_serialization(self, send_payload):

        class tmp(object):
            @property
            def __class__(self):
                foo()

        try:
            t = tmp()
            raise Exception('trigger_serialize')
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

    @mock.patch('rollbar.send_payload')
    def test_args_lambda_no_args(self, send_payload):

        _raise = lambda: foo()

        try:
            _raise()
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertNotIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('locals', payload['data']['body']['trace']['frames'][-1])

    @mock.patch('rollbar.send_payload')
    def test_args_lambda_with_args(self, send_payload):

        _raise = lambda arg1, arg2: foo(arg1, arg2)

        try:
            _raise('arg1-value', 'arg2-value')
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        self.assertEqual(2, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual('arg1', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertEqual('arg1-value', payload['data']['body']['trace']['frames'][-1]['locals']['arg1'])
        self.assertEqual('arg2', payload['data']['body']['trace']['frames'][-1]['argspec'][1])
        self.assertEqual('arg2-value', payload['data']['body']['trace']['frames'][-1]['locals']['arg2'])

    @mock.patch('rollbar.send_payload')
    def test_args_lambda_with_defaults(self, send_payload):

        _raise = lambda arg1='default': foo(arg1)

        try:
            _raise(arg1='arg1-value')
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        # NOTE(cory): Lambdas are a bit strange. We treat default values for lambda args
        #             as positional.
        self.assertEqual(1, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual('arg1', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertEqual('arg1-value', payload['data']['body']['trace']['frames'][-1]['locals']['arg1'])

    @mock.patch('rollbar.send_payload')
    def test_args_lambda_with_star_args(self, send_payload):

        _raise = lambda *args: foo(arg1)

        try:
            _raise('arg1-value')
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertNotIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        varargs = payload['data']['body']['trace']['frames'][-1]['varargspec']

        self.assertEqual(1, len(payload['data']['body']['trace']['frames'][-1]['locals'][varargs]))
        self.assertRegex(payload['data']['body']['trace']['frames'][-1]['locals'][varargs][0], r'\*+')

    @mock.patch('rollbar.send_payload')
    def test_args_lambda_with_star_args_and_args(self, send_payload):

        _raise = lambda arg1, *args: foo(arg1)

        try:
            _raise('arg1-value', 1, 2)
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        varargs = payload['data']['body']['trace']['frames'][-1]['varargspec']

        self.assertEqual(1, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual('arg1', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertEqual('arg1-value', payload['data']['body']['trace']['frames'][-1]['locals']['arg1'])

        self.assertEqual(2, len(payload['data']['body']['trace']['frames'][-1]['locals'][varargs]))
        self.assertRegex(payload['data']['body']['trace']['frames'][-1]['locals'][varargs][0], r'\*+')
        self.assertRegex(payload['data']['body']['trace']['frames'][-1]['locals'][varargs][1], r'\*+')

    @mock.patch('rollbar.send_payload')
    def test_args_lambda_with_kwargs(self, send_payload):

        _raise = lambda **kwargs: foo(arg1)

        try:
            _raise(arg1='arg1-value', arg2=2)
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertNotIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        keywords = payload['data']['body']['trace']['frames'][-1]['keywordspec']

        self.assertEqual(2, len(payload['data']['body']['trace']['frames'][-1]['locals'][keywords]))
        self.assertEqual('arg1-value', payload['data']['body']['trace']['frames'][-1]['locals'][keywords]['arg1'])
        self.assertEqual(2, payload['data']['body']['trace']['frames'][-1]['locals'][keywords]['arg2'])

    @mock.patch('rollbar.send_payload')
    def test_args_lambda_with_kwargs_and_args(self, send_payload):

        _raise = lambda arg1, arg2, **kwargs: foo(arg1)

        try:
            _raise('a1', 'a2', arg3='arg3-value', arg4=2)
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        keywords = payload['data']['body']['trace']['frames'][-1]['keywordspec']

        self.assertEqual(2, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual('arg1', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertEqual('arg2', payload['data']['body']['trace']['frames'][-1]['argspec'][1])
        self.assertEqual('a1', payload['data']['body']['trace']['frames'][-1]['locals']['arg1'])
        self.assertEqual('a2', payload['data']['body']['trace']['frames'][-1]['locals']['arg2'])

        self.assertEqual(2, len(payload['data']['body']['trace']['frames'][-1]['locals'][keywords]))
        self.assertEqual('arg3-value', payload['data']['body']['trace']['frames'][-1]['locals'][keywords]['arg3'])
        self.assertEqual(2, payload['data']['body']['trace']['frames'][-1]['locals'][keywords]['arg4'])

    @mock.patch('rollbar.send_payload')
    def test_args_lambda_with_kwargs_and_args_and_defaults(self, send_payload):

        _raise = lambda arg1, arg2, arg3='default-value', **kwargs: foo(arg1)

        try:
            _raise('a1', 'a2', arg3='arg3-value', arg4=2)
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        keywords = payload['data']['body']['trace']['frames'][-1]['keywordspec']

        # NOTE(cory): again, default values are strange for lambdas and we include them as
        #             positional args.
        self.assertEqual(3, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual('arg1', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertEqual('arg2', payload['data']['body']['trace']['frames'][-1]['argspec'][1])
        self.assertEqual('arg3', payload['data']['body']['trace']['frames'][-1]['argspec'][2])
        self.assertEqual('a1', payload['data']['body']['trace']['frames'][-1]['locals']['arg1'])
        self.assertEqual('a2', payload['data']['body']['trace']['frames'][-1]['locals']['arg2'])
        self.assertEqual('arg3-value', payload['data']['body']['trace']['frames'][-1]['locals']['arg3'])

        self.assertEqual(1, len(payload['data']['body']['trace']['frames'][-1]['locals'][keywords]))
        self.assertEqual(2, payload['data']['body']['trace']['frames'][-1]['locals'][keywords]['arg4'])

    @mock.patch('rollbar.send_payload')
    def test_args_generators(self, send_payload):

        def _raise(arg1):
            for i in range(2):
                if i > 0:
                    raise Exception()
                else:
                    yield i

        try:
            l = list(_raise('hello world'))
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        self.assertEqual(1, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual('arg1', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertEqual('hello world', payload['data']['body']['trace']['frames'][-1]['locals']['arg1'])

    @mock.patch('rollbar.send_payload')
    def test_anonymous_tuple_args(self, send_payload):

        # Only run this test on Python versions that support it
        if not _anonymous_tuple_func:
            return

        try:
            _anonymous_tuple_func((1, (2, 3), 4))
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        self.assertEqual(4, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual(1, payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertEqual(2, payload['data']['body']['trace']['frames'][-1]['argspec'][1])
        self.assertEqual(3, payload['data']['body']['trace']['frames'][-1]['argspec'][2])
        self.assertEqual(4, payload['data']['body']['trace']['frames'][-1]['argspec'][3])
        self.assertEqual(10, payload['data']['body']['trace']['frames'][-1]['locals']['ret'])

    @mock.patch('rollbar.send_payload')
    def test_scrub_defaults(self, send_payload):

        def _raise(password='sensitive', clear='text'):
            headers = { 'Authorization': 'bearer 123' }

            raise Exception()

        try:
            _raise()
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('kwargs', payload['data']['body']['trace']['frames'][-1]['locals'])

        self.assertEqual(2, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual('password', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertRegex(payload['data']['body']['trace']['frames'][-1]['locals']['password'], r'\*+')
        self.assertRegex(payload['data']['body']['trace']['frames'][-1]['locals']['headers']['Authorization'], r'\*+')
        self.assertEqual('clear', payload['data']['body']['trace']['frames'][-1]['argspec'][1])
        self.assertEqual('text', payload['data']['body']['trace']['frames'][-1]['locals']['clear'])

    @mock.patch('rollbar.send_payload')
    def test_dont_scrub_star_args(self, send_payload):
        rollbar.SETTINGS['locals']['scrub_varargs'] = False

        def _raise(*args):
            raise Exception()

        try:
            _raise('sensitive', 'text')
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertNotIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('locals', payload['data']['body']['trace']['frames'][-1])

        varargspec = payload['data']['body']['trace']['frames'][-1]['varargspec']

        self.assertEqual(2, len(payload['data']['body']['trace']['frames'][-1]['locals'][varargspec]))
        self.assertEqual(payload['data']['body']['trace']['frames'][-1]['locals'][varargspec][0], 'sensitive')
        self.assertEqual(payload['data']['body']['trace']['frames'][-1]['locals'][varargspec][1], 'text')

    @mock.patch('rollbar.send_payload')
    def test_scrub_kwargs(self, send_payload):

        def _raise(**kwargs):
            raise Exception()

        try:
            _raise(password='sensitive', clear='text')
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertNotIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        keywords = payload['data']['body']['trace']['frames'][-1]['keywordspec']

        self.assertEqual(2, len(payload['data']['body']['trace']['frames'][-1]['locals'][keywords]))
        self.assertIn('password', payload['data']['body']['trace']['frames'][-1]['locals'][keywords])
        self.assertRegex(payload['data']['body']['trace']['frames'][-1]['locals'][keywords]['password'], r'\*+')
        self.assertIn('clear', payload['data']['body']['trace']['frames'][-1]['locals'][keywords])
        self.assertEqual('text', payload['data']['body']['trace']['frames'][-1]['locals'][keywords]['clear'])

    @mock.patch('rollbar.send_payload')
    def test_scrub_locals(self, send_payload):
        invalid_b64 = b'CuX2JKuXuLVtJ6l1s7DeeQ=='
        invalid = base64.b64decode(invalid_b64)

        def _raise():
            # Make sure that the _invalid local variable makes its
            # way into the payload even if its value cannot be serialized
            # properly.
            _invalid = invalid

            # Make sure the Password field gets scrubbed even though its
            # original value could not be serialized properly.
            Password = invalid

            password = 'sensitive'
            raise Exception((_invalid, Password, password))

        try:
            _raise()
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertRegex(payload['data']['body']['trace']['frames'][-1]['locals']['password'], r'\*+')
        self.assertRegex(payload['data']['body']['trace']['frames'][-1]['locals']['Password'], r'\*+')
        self.assertIn('_invalid', payload['data']['body']['trace']['frames'][-1]['locals'])

        undecodable_message = '<Undecodable type:(%s) base64:(%s)>' % ('bytes', base64.b64encode(invalid).decode('ascii'))
        self.assertEqual(undecodable_message, payload['data']['body']['trace']['frames'][-1]['locals']['_invalid'])

    @mock.patch('rollbar.send_payload')
    def test_scrub_nans(self, send_payload):
        def _raise():
            infinity = float('Inf')
            negative_infinity = float('-Inf')
            not_a_number = float('NaN')
            raise Exception()

        try:
            _raise()
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertEqual('<Infinity>', payload['data']['body']['trace']['frames'][-1]['locals']['infinity'])
        self.assertEqual('<NegativeInfinity>', payload['data']['body']['trace']['frames'][-1]['locals']['negative_infinity'])
        self.assertEqual('<NaN>', payload['data']['body']['trace']['frames'][-1]['locals']['not_a_number'])

    @mock.patch('rollbar.send_payload')
    def test_scrub_self_referencing(self, send_payload):
        def _raise(obj):
            raise Exception()

        try:
            obj = {'x': 42.3}
            obj['child'] = {
                'parent': obj
            }

            # NOTE(cory): We copy the dict here so that we don't produce a circular reference
            # from the _rase() args.
            _raise(dict(obj))
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertTrue(
            (isinstance(payload['data']['body']['trace']['frames'][-1]['locals']['obj'], dict) and
             'child' in payload['data']['body']['trace']['frames'][-1]['locals']['obj'])

             or

            (isinstance(payload['data']['body']['trace']['frames'][-1]['locals']['obj'], string_types) and
             payload['data']['body']['trace']['frames'][-1]['locals']['obj'].startswith('<CircularReference'))
        )

        self.assertTrue(
            (isinstance(payload['data']['body']['trace']['frames'][-1]['locals']['obj'], dict) and
             'x' in payload['data']['body']['trace']['frames'][-1]['locals']['obj'] and
             payload['data']['body']['trace']['frames'][-1]['locals']['obj']['x'] == 42.3)

             or

            (isinstance(payload['data']['body']['trace']['frames'][-1]['locals']['obj'], string_types) and
             payload['data']['body']['trace']['frames'][-1]['locals']['obj'].startswith('<CircularReference'))
        )

    @mock.patch('rollbar.send_payload')
    def test_scrub_local_ref(self, send_payload):
        """
        NOTE(cory): This test checks to make sure that we do not scrub a local variable that is a reference
                    to a parameter that is scrubbed.
                    Ideally we would be able to scrub 'copy' as well since we know that it has the same
                    value as a field that was scrubbed.
        """
        def _raise(password='sensitive'):
            copy = password
            raise Exception()

        try:
            _raise()
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertEqual('sensitive', payload['data']['body']['trace']['frames'][-1]['locals']['copy'])

    @mock.patch('rollbar.send_payload')
    def test_large_arg_val(self, send_payload):

        def _raise(large):
            raise Exception()

        try:
            large = ''.join(['#'] * 200)
            _raise(large)
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        self.assertEqual(1, len(payload['data']['body']['trace']['frames'][-1]['argspec']))
        self.assertEqual('large', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertEqual("'###############################################...################################################'",
                         payload['data']['body']['trace']['frames'][-1]['locals']['large'])

    @mock.patch('rollbar.send_payload')
    def test_long_list_arg_val(self, send_payload):

        def _raise(large):
            raise Exception()

        try:
            xlarge = ['hi' for _ in range(30)]
            # NOTE(cory): We copy the list here so that the local variables from
            # this frame are not referenced directly by the frame from _raise()
            # call above. If we didn't copy this list, Rollbar would report a
            # circular reference for the args on _raise().
            _raise([str(x) for x in xlarge])
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        self.assertEqual(1, len(payload['data']['body']['trace']['frames'][-1]['argspec']))

        self.assertEqual('large', payload['data']['body']['trace']['frames'][-1]['argspec'][0])
        self.assertTrue(
            ("['hi', 'hi', 'hi', 'hi', 'hi', 'hi', 'hi', 'hi', 'hi', 'hi', ...]" ==
                payload['data']['body']['trace']['frames'][-1]['argspec'][0])

            or

            ("['hi', 'hi', 'hi', 'hi', 'hi', 'hi', 'hi', 'hi', 'hi', 'hi', ...]" ==
                    payload['data']['body']['trace']['frames'][0]['locals']['xlarge']))


    @mock.patch('rollbar.send_payload')
    def test_last_frame_has_locals(self, send_payload):

        def _raise():
            some_var = 'some value'
            raise Exception()

        try:
            _raise()
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        self.assertNotIn('argspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('varargspec', payload['data']['body']['trace']['frames'][-1])
        self.assertNotIn('keywordspec', payload['data']['body']['trace']['frames'][-1])

        self.assertIn('locals', payload['data']['body']['trace']['frames'][-1])
        self.assertIn('some_var', payload['data']['body']['trace']['frames'][-1]['locals'])
        self.assertEqual("some value",
                         payload['data']['body']['trace']['frames'][-1]['locals']['some_var'])


    @mock.patch('rollbar.send_payload')
    def test_all_project_frames_have_locals(self, send_payload):

        prev_root = rollbar.SETTINGS['root']
        rollbar.SETTINGS['root'] = __file__.rstrip('pyc')
        try:
            step1()
        except:
            rollbar.report_exc_info()
        finally:
            rollbar.SETTINGS['root'] = prev_root

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]
        for frame in payload['data']['body']['trace']['frames']:
            self.assertIn('locals', frame)


    @mock.patch('rollbar.send_payload')
    def test_only_last_frame_has_locals(self, send_payload):

        prev_root = rollbar.SETTINGS['root']
        rollbar.SETTINGS['root'] = 'dummy'
        try:
            step1()
        except:
            rollbar.report_exc_info()
        finally:
            rollbar.SETTINGS['root'] = prev_root

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        num_frames = len(payload['data']['body']['trace']['frames'])
        for i, frame in enumerate(payload['data']['body']['trace']['frames']):
            if i < num_frames - 1:
                self.assertNotIn('locals', frame)
            else:
                self.assertIn('locals', frame)


    @mock.patch('rollbar.send_payload')
    def test_modify_arg(self, send_payload):
        # Record locals for all frames
        prev_root = rollbar.SETTINGS['root']
        rollbar.SETTINGS['root'] = __file__.rstrip('pyc')
        try:
            called_with('original value')
        except:
            rollbar.report_exc_info()
        finally:
            rollbar.SETTINGS['root'] = prev_root

        self.assertEqual(send_payload.called, True)

        payload = send_payload.call_args[0][0]

        frames = payload['data']['body']['trace']['frames']
        called_with_frame = frames[1]

        self.assertEqual('arg1', called_with_frame['argspec'][0])
        self.assertEqual('changed', called_with_frame['locals']['arg1'])

    @mock.patch('rollbar.send_payload')
    def test_unicode_exc_info(self, send_payload):
        message = '\u221a'

        try:
            raise Exception(message)
        except:
            rollbar.report_exc_info()

        self.assertEqual(send_payload.called, True)
        payload = send_payload.call_args[0][0]
        self.assertEqual(payload['data']['body']['trace']['exception']['message'], message)

    @mock.patch('rollbar.lib.transport.post', side_effect=lambda *args, **kw: MockResponse({'status': 'OK'}, 200))
    def test_serialize_and_send_payload(self, post=None):
        invalid_b64 = b'CuX2JKuXuLVtJ6l1s7DeeQ=='
        invalid = base64.b64decode(invalid_b64)

        def _raise():
            # Make sure that the _invalid local variable makes its
            # way into the payload even if its value cannot be serialized
            # properly.
            _invalid = invalid

            # Make sure the Password field gets scrubbed even though its
            # original value could not be serialized properly.
            Password = invalid

            password = 'sensitive'
            raise Exception('bug bug')

        try:
            _raise()
        except:
            rollbar.report_exc_info()

        self.assertEqual(post.called, True)
        payload_data = post.call_args[1]['data']
        self.assertIsInstance(payload_data, str)
        self.assertIn('bug bug', payload_data)

        try:
            post.call_args[1]['data']
        except:
            self.assertTrue(False)

    def test_scrub_webob_request_data(self):
        rollbar._initialized = False
        rollbar.init(_test_access_token, locals={'enabled': True}, dummy_key='asdf', handler='blocking', timeout=12345,
            scrub_fields=rollbar.SETTINGS['scrub_fields'] + ['token', 'secret', 'cookies', 'authorization'])

        import webob
        request = webob.Request.blank('/the/path?q=hello&password=hunter2',
                                      base_url='http://example.com',
                                      headers={
                                          'X-Real-Ip': '5.6.7.8',
                                          'Cookies': 'name=value; password=hash;',
                                          'Authorization': 'I am from NSA'
                                      },
                                      POST='foo=bar&confirm_password=hunter3&token=secret')

        unscrubbed = rollbar._build_webob_request_data(request)
        self.assertEqual(unscrubbed['url'], 'http://example.com/the/path?q=hello&password=hunter2')
        self.assertEqual(unscrubbed['user_ip'], '5.6.7.8')
        self.assertDictEqual(unscrubbed['GET'], {'q': 'hello', 'password': 'hunter2'})
        self.assertDictEqual(unscrubbed['POST'], {'foo': 'bar', 'confirm_password': 'hunter3', 'token': 'secret'})
        self.assertEqual('5.6.7.8', unscrubbed['headers']['X-Real-Ip'])
        self.assertEqual('name=value; password=hash;', unscrubbed['headers']['Cookies'])
        self.assertEqual('I am from NSA', unscrubbed['headers']['Authorization'])

        scrubbed = rollbar._transform(unscrubbed)
        self.assertRegex(scrubbed['url'], r'http://example.com/the/path\?(q=hello&password=-+)|(password=-+&q=hello)')

        self.assertEqual(scrubbed['GET']['q'], 'hello')
        self.assertRegex(scrubbed['GET']['password'], r'\*+')

        self.assertEqual(scrubbed['POST']['foo'], 'bar')
        self.assertRegex(scrubbed['POST']['confirm_password'], r'\*+')
        self.assertRegex(scrubbed['POST']['token'], r'\*+')

        self.assertEqual('5.6.7.8', scrubbed['headers']['X-Real-Ip'])

        self.assertRegex(scrubbed['headers']['Cookies'], r'\*+')
        self.assertRegex(scrubbed['headers']['Authorization'], r'\*+')

    def test_filter_ip_no_user_ip(self):
        request_data = {'something': 'but no ip'}
        rollbar._filter_ip(request_data, False)
        self.assertNotIn('user_ip', request_data)

    def test_filter_ip_capture_true(self):
        ip = '123.32.394.99'
        request_data = {'user_ip': ip}
        rollbar._filter_ip(request_data, True)
        self.assertEqual(ip, request_data['user_ip'])

    def test_filter_ip_anonymize(self):
        ip = '123.32.394.99'
        request_data = {'user_ip': ip}
        rollbar._filter_ip(request_data, rollbar.ANONYMIZE)
        self.assertNotEqual(ip, request_data['user_ip'])
        self.assertNotEqual(None, request_data['user_ip'])

    def test_filter_ip_capture_false(self):
        ip = '123.32.394.99'
        request_data = {'user_ip': ip}
        rollbar._filter_ip(request_data, False)
        self.assertNotEqual(ip, request_data['user_ip'])
        self.assertEqual(None, request_data['user_ip'])

    def test_filter_ip_ipv6_capture_false(self):
        ip = '2607:f0d0:1002:51::4'
        request_data = {'user_ip': ip}
        rollbar._filter_ip(request_data, False)
        self.assertNotEqual(ip, request_data['user_ip'])
        self.assertEqual(None, request_data['user_ip'])

    def test_filter_ip_anonymize_ipv6(self):
        ips = [
            'FE80:0000:0000:0000:0202:B3FF:FE1E:8329',
            'FE80::0202:B3FF:FE1E:8329',
            '2607:f0d0:1002:51::4',
        ]
        for ip in ips:
            request_data = {'user_ip': ip}
            rollbar._filter_ip(request_data, rollbar.ANONYMIZE)
            self.assertNotEqual(ip, request_data['user_ip'])
            self.assertNotEqual(None, request_data['user_ip'])

    def test_starlette_extract_user_ip_from_client_host(self):
        try:
            from starlette.requests import Request
        except ImportError:
            self.skipTest('Requires Starlette package')

        client_host = ('127.0.0.1', 1453)
        ip_forwarded_for = b'192.168.10.10'
        ip_real_ip = b'1.2.3.4'
        scope = {
            'type': 'http',
            'client': client_host,
            'headers': [
                (b'x-forwarded-for', ip_forwarded_for),
                (b'x-real-ip', ip_real_ip),
            ],
        }
        request = Request(scope)

        user_ip = rollbar._starlette_extract_user_ip(request)

        self.assertEqual(user_ip, client_host[0])

    def test_starlette_extract_user_ip_from_headers(self):
        try:
            from starlette.requests import Request
        except ImportError:
            self.skipTest('Requires Starlette package')

        ip_forwarded_for = b'192.168.10.10'
        ip_real_ip = b'1.2.3.4'

        # Headers contain only X-Forwarded-For
        scope = {'type': 'http', 'headers': [(b'x-forwarded-for', ip_forwarded_for)]}
        request = Request(scope)
        user_ip = rollbar._starlette_extract_user_ip(request)
        self.assertEqual(user_ip, ip_forwarded_for.decode())

        # Headers contain only X-Real-Ip
        scope = {'type': 'http', 'headers': [(b'x-real-ip', ip_real_ip)]}
        request = Request(scope)
        user_ip = rollbar._starlette_extract_user_ip(request)
        self.assertEqual(user_ip, ip_real_ip.decode())

        # Headers contain both X-Forwarded-For and X-Real-Ip
        scope = {
            'type': 'http',
            'headers': [
                (b'x-forwarded-for', ip_forwarded_for),
                (b'x-real-ip', ip_real_ip),
            ],
        }
        request = Request(scope)
        user_ip = rollbar._starlette_extract_user_ip(request)
        self.assertEqual(user_ip, ip_forwarded_for.decode())

### Helpers

def step1():
    val1 = 1
    step2()


def step2():
    val2 = 2
    raise Exception()


def called_with(arg1):
    arg1 = 'changed'
    step1()


class CauseException(Exception):
    pass


class MockResponse:
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code

    @property
    def content(self):
        return json.dumps(self.json_data)

    def json(self):
        return self.json_data


class MockRawResponse:
    def __init__(self, data, status_code):
        self.data = data
        self.status_code = status_code

    @property
    def content(self):
        return self.data

    def json(self):
        return self.data


class MockLambdaContext(object):
    def __init__(self, x):
        self.function_name = 1
        self.function_version = 2
        self.invoked_function_arn = 3
        self.aws_request_id = 4
        self.x = x

    def get_remaining_time_in_millis(self):
        42


def fastapi_add_route_with_request_param(app, endpoint, path, request_param):
    from fastapi import Request

    endpoint.__annotations__[request_param] = Request

    return app.get(path)(endpoint)

if __name__ == '__main__':
    unittest.main()



from rollbar.lib import events
from rollbar.lib.filters.basic import filter_rollbar_ignored_exceptions, filter_by_level


def add_builtin_filters(settings):
    # exc_info filters
    events.add_exception_info_handler(filter_rollbar_ignored_exceptions)
    events.add_exception_info_handler(filter_by_level)

    # message filters
    events.add_message_handler(filter_by_level)



from __future__ import absolute_import
from __future__ import unicode_literals

import copy
import functools
import inspect
import json
import logging
import os
import socket
import sys
import threading
import time
import traceback
import types
import uuid
import wsgiref.util
import warnings
import queue
from urllib.parse import parse_qs, urljoin

import requests

from rollbar.lib import events, filters, dict_merge, transport, defaultJSONEncode


__version__ = '1.0.0b1'
__log_name__ = 'rollbar'
log = logging.getLogger(__log_name__)


# import request objects from various frameworks, if available
try:
    from webob import BaseRequest as WebobBaseRequest
except ImportError:
    WebobBaseRequest = None

try:
    from django.core.exceptions import ImproperlyConfigured
except ImportError:
    DjangoHttpRequest = None
    RestFrameworkRequest = None

else:
    try:
        from django.http import HttpRequest as DjangoHttpRequest
    except (ImportError, ImproperlyConfigured):
        DjangoHttpRequest = None

    try:
        from rest_framework.request import Request as RestFrameworkRequest
    except (ImportError, ImproperlyConfigured):
        RestFrameworkRequest = None

    del ImproperlyConfigured

try:
    from werkzeug.wrappers import Request as WerkzeugRequest
except (ImportError, SyntaxError):
    WerkzeugRequest = None

try:
    from werkzeug.local import LocalProxy as WerkzeugLocalProxy
except (ImportError, SyntaxError):
    WerkzeugLocalProxy = None

try:
    from tornado.httpserver import HTTPRequest as TornadoRequest
except ImportError:
    TornadoRequest = None

try:
    from bottle import BaseRequest as BottleRequest
except ImportError:
    BottleRequest = None

try:
    from sanic.request import Request as SanicRequest
except ImportError:
    SanicRequest = None

try:
    from google.appengine.api.urlfetch import fetch as AppEngineFetch
except (ImportError, KeyError):
    AppEngineFetch = None

try:
    from starlette.requests import Request as StarletteRequest
except ImportError:
    StarletteRequest = None

try:
    from fastapi.requests import Request as FastAPIRequest
except ImportError:
    FastAPIRequest = None

try:
    import httpx
except ImportError:
    httpx = None

AsyncHTTPClient = httpx

def passthrough_decorator(func):
    def wrap(*args, **kwargs):
        return func(*args, **kwargs)
    return wrap

try:
    from tornado.httpclient import AsyncHTTPClient as TornadoAsyncHTTPClient
except ImportError:
    TornadoAsyncHTTPClient = None

try:
    import treq
    from twisted.python import log as twisted_log
    from twisted.web.iweb import IPolicyForHTTPS
    from twisted.web.client import BrowserLikePolicyForHTTPS, Agent
    from twisted.internet.ssl import CertificateOptions
    from twisted.internet import task, defer, ssl, reactor
    from zope.interface import implementer

    @implementer(IPolicyForHTTPS)
    class VerifyHTTPS(object):
        def __init__(self):
            # by default, handle requests like a browser would
            self.default_policy = BrowserLikePolicyForHTTPS()

        def creatorForNetloc(self, hostname, port):
            # check if the hostname is in the the whitelist, otherwise return the default policy
            if not SETTINGS['verify_https']:
                return ssl.CertificateOptions(verify=False)
            return self.default_policy.creatorForNetloc(hostname, port)

    def log_handler(event):
        """
        Default uncaught error handler
        """
        try:
            if not event.get('isError') or 'failure' not in event:
                return

            err = event['failure']

            # Don't report Rollbar internal errors to ourselves
            if issubclass(err.type, ApiException):
                log.error('Rollbar internal error: %s', err.value)
            else:
                report_exc_info((err.type, err.value, err.getTracebackObject()))
        except:
            log.exception('Error while reporting to Rollbar')

    # Add Rollbar as a log handler which will report uncaught errors
    twisted_log.addObserver(log_handler)


except ImportError:
    treq = None

try:
    from falcon import Request as FalconRequest
except ImportError:
    FalconRequest = None


def get_request():
    """
    Get the current request object. Implementation varies on
    library support. Modified below when we know which framework
    is being used.
    """

    # TODO(cory): add in a generic _get_locals_request() which
    # will iterate up through the call stack and look for a variable
    # that appears to be valid request object.
    for fn in (_get_fastapi_request,
               _get_starlette_request,
               _get_bottle_request,
               _get_flask_request,
               _get_pyramid_request,
               _get_pylons_request):
        try:
            req = fn()
            if req is not None:
                return req
        except:
            pass

    return None


def _get_bottle_request():
    if BottleRequest is None:
        return None
    from bottle import request
    return request


def _get_flask_request():
    if WerkzeugRequest is None:
        return None
    from flask import request
    return request


def _get_pyramid_request():
    if WebobBaseRequest is None:
        return None
    from pyramid.threadlocal import get_current_request
    return get_current_request()


def _get_pylons_request():
    if WebobBaseRequest is None:
        return None
    from pylons import request
    return request


def _get_starlette_request():
    # Do not modify the returned object

    if StarletteRequest is None:
        return None

    from rollbar.contrib.starlette import get_current_request
    return get_current_request()


def _get_fastapi_request():
    # Do not modify the returned object

    if FastAPIRequest is None:
        return None

    from rollbar.contrib.fastapi import get_current_request
    return get_current_request()


BASE_DATA_HOOK = None

agent_log = None

VERSION = __version__
DEFAULT_ENDPOINT = 'https://api.rollbar.com/api/1/'
DEFAULT_TIMEOUT = 3
ANONYMIZE = 'anonymize'

DEFAULT_LOCALS_SIZES = {
    'maxlevel': 5,
    'maxdict': 10,
    'maxlist': 10,
    'maxtuple': 10,
    'maxset': 10,
    'maxfrozenset': 10,
    'maxdeque': 10,
    'maxarray': 10,
    'maxstring': 100,
    'maxlong': 40,
    'maxother': 100,
}

# configuration settings
# configure by calling init() or overriding directly
SETTINGS = {
    'access_token': None,
    'enabled': True,
    'environment': 'production',
    'exception_level_filters': [],
    'root': None,  # root path to your code
    'branch': None,  # git branch name
    'code_version': None,
    # 'blocking', 'thread' (default), 'async', 'agent', 'tornado', 'gae', 'twisted', 'httpx' or 'thread_pool'
    # 'async' requires Python 3.4 or higher.
    # 'httpx' requires Python 3.7 or higher.
    # 'thread_pool' requires Python 3.2 or higher.
    'handler': 'default',
    'thread_pool_workers': None,
    'endpoint': DEFAULT_ENDPOINT,
    'timeout': DEFAULT_TIMEOUT,
    'agent.log_file': 'log.rollbar',
    'scrub_fields': [
        'pw',
        'passwd',
        'password',
        'secret',
        'confirm_password',
        'confirmPassword',
        'password_confirmation',
        'passwordConfirmation',
        'access_token',
        'accessToken',
        'auth',
        'authentication',
        'authorization',
    ],
    'url_fields': ['url', 'link', 'href'],
    'notifier': {
        'name': 'pyrollbar',
        'version': VERSION
    },
    'allow_logging_basic_config': True,  # set to False to avoid a call to logging.basicConfig()
    'locals': {
        'enabled': True,
        'safe_repr': True,
        'scrub_varargs': True,
        'sizes': DEFAULT_LOCALS_SIZES,
        'safelisted_types': [],
        'whitelisted_types': []
    },
    'verify_https': True,
    'shortener_keys': [],
    'suppress_reinit_warning': False,
    'capture_email': False,
    'capture_username': False,
    'capture_ip': True,
    'log_all_rate_limited_items': True,
    'http_proxy': None,
    'http_proxy_user': None,
    'http_proxy_password': None,
    'include_request_body': False,
    'request_pool_connections': None,
    'request_pool_maxsize': None,
    'request_max_retries': None,
    'batch_transforms': False,
}

_CURRENT_LAMBDA_CONTEXT = None
_LAST_RESPONSE_STATUS = None

# Set in init()
_transforms = []
_serialize_transform = None

_initialized = False

from rollbar.lib.transforms.scrub_redact import REDACT_REF

from rollbar.lib import transforms
from rollbar.lib import type_info
from rollbar.lib.transforms.scrub import ScrubTransform
from rollbar.lib.transforms.scruburl import ScrubUrlTransform
from rollbar.lib.transforms.scrub_redact import ScrubRedactTransform
from rollbar.lib.transforms.serializable import SerializableTransform
from rollbar.lib.transforms.shortener import ShortenerTransform
from rollbar.lib.transforms.batched import BatchedTransform


## public api

def init(access_token, environment='production', scrub_fields=None, url_fields=None, **kw):
    """
    Saves configuration variables in this module's SETTINGS.

    access_token: project access token. Get this from the Rollbar UI:
                  - click "Settings" in the top nav
                  - click "Projects" in the left nav
                  - copy-paste the appropriate token.
    environment: environment name. Can be any string; suggestions: 'production', 'development',
                 'staging', 'yourname'
    **kw: provided keyword arguments will override keys in SETTINGS.
    """
    global SETTINGS, agent_log, _initialized, _transforms, _serialize_transform, _threads

    if scrub_fields is not None:
       SETTINGS['scrub_fields'] = list(scrub_fields)
    if url_fields is not None:
       SETTINGS['url_fields'] = list(url_fields)

    # Merge the extra config settings into SETTINGS
    SETTINGS = dict_merge(SETTINGS, kw)
    if _initialized:
        # NOTE: Temp solution to not being able to re-init.
        # New versions of pyrollbar will support re-initialization
        # via the (not-yet-implemented) configure() method.
        if not SETTINGS.get('suppress_reinit_warning'):
            log.warning('Rollbar already initialized. Ignoring re-init.')
        return

    SETTINGS['access_token'] = access_token
    SETTINGS['environment'] = environment
    _configure_transport(**SETTINGS)

    if SETTINGS.get('allow_logging_basic_config'):
        logging.basicConfig()

    if SETTINGS.get('handler') == 'agent':
        agent_log = _create_agent_log()
    elif SETTINGS.get('handler') == 'thread_pool':
        from rollbar.lib.thread_pool import init_pool
        init_pool(SETTINGS.get('thread_pool_workers', None))

    if not SETTINGS['locals']['safelisted_types'] and SETTINGS['locals']['whitelisted_types']:
        warnings.warn('whitelisted_types deprecated use safelisted_types instead', DeprecationWarning)
        SETTINGS['locals']['safelisted_types'] = SETTINGS['locals']['whitelisted_types']

    # We will perform these transforms in order:
    # 1. Serialize the payload to be all python built-in objects
    # 2. Scrub the payloads based on the key suffixes in SETTINGS['scrub_fields']
    # 3. Scrub URLs in the payload for keys that end with 'url'
    # 4. Optional - If local variable gathering is enabled, transform the
    #       trace frame values using the ShortReprTransform.
    _serialize_transform = SerializableTransform(safe_repr=SETTINGS['locals']['safe_repr'],
                                                 safelist_types=SETTINGS['locals']['safelisted_types'])
    _transforms = [
        ScrubRedactTransform(),
        _serialize_transform,
        ScrubTransform(suffixes=[(field,) for field in SETTINGS['scrub_fields']], redact_char='*'),
        ScrubUrlTransform(suffixes=[(field,) for field in SETTINGS['url_fields']], params_to_scrub=SETTINGS['scrub_fields'])
    ]

    # A list of key prefixes to apply our shortener transform to. The request
    # being included in the body key is old behavior and is being retained for
    # backwards compatibility.
    shortener_keys = [
        ('request', 'POST'),
        ('request', 'json'),
        ('body', 'request', 'POST'),
        ('body', 'request', 'json'),
    ]

    if SETTINGS['locals']['enabled']:
        for prefix in (('body', 'trace'), ('body', 'trace_chain', '*')):
            shortener_keys.append(prefix + ('frames', '*', 'code'))
            shortener_keys.append(prefix + ('frames', '*', 'args', '*'))
            shortener_keys.append(prefix + ('frames', '*', 'kwargs', '*'))
            shortener_keys.append(prefix + ('frames', '*', 'locals', '*'))

    shortener_keys.extend(SETTINGS['shortener_keys'])

    shortener = ShortenerTransform(safe_repr=SETTINGS['locals']['safe_repr'],
                                   keys=shortener_keys,
                                   **SETTINGS['locals']['sizes'])
    _transforms.append(shortener)
    _threads = queue.Queue()
    events.reset()
    filters.add_builtin_filters(SETTINGS)

    _initialized = True


def _configure_transport(**kw):
    configuration = _requests_configuration(**kw)
    transport.configure_pool(**configuration)


def _requests_configuration(**kw):
    keys = {
        'request_pool_connections': 'pool_connections',
        'request_pool_maxsize': 'pool_maxsize',
        'request_max_retries': 'max_retries',
    }
    return {keys[k]: kw.get(k, None) for k in keys}


def lambda_function(f):
    """
    Decorator for making error handling on AWS Lambda easier
    """
    @functools.wraps(f)
    def wrapper(event, context):
        global _CURRENT_LAMBDA_CONTEXT
        _CURRENT_LAMBDA_CONTEXT = context
        try:
            result = f(event, context)
            return wait(lambda: result)
        except:
            cls, exc, trace = sys.exc_info()
            report_exc_info((cls, exc, trace.tb_next))
            wait()
            raise
    return wrapper


def report_exc_info(exc_info=None, request=None, extra_data=None, payload_data=None, level=None, **kw):
    """
    Reports an exception to Rollbar, using exc_info (from calling sys.exc_info())

    exc_info: optional, should be the result of calling sys.exc_info(). If omitted, sys.exc_info() will be called here.
    request: optional, a WebOb, Werkzeug-based or Sanic request object.
    extra_data: optional, will be included in the 'custom' section of the payload
    payload_data: optional, dict that will override values in the final payload
                  (e.g. 'level' or 'fingerprint')
    kw: provided for legacy purposes; unused.

    Example usage:

    rollbar.init(access_token='YOUR_PROJECT_ACCESS_TOKEN')
    try:
        do_something()
    except:
        rollbar.report_exc_info(sys.exc_info(), request, {'foo': 'bar'}, {'level': 'warning'})
    """
    if exc_info is None:
        exc_info = sys.exc_info()

    try:
        return _report_exc_info(exc_info, request, extra_data, payload_data, level=level)
    except Exception as e:
        log.exception("Exception while reporting exc_info to Rollbar. %r", e)


def report_message(message, level='error', request=None, extra_data=None, payload_data=None):
    """
    Reports an arbitrary string message to Rollbar.

    message: the string body of the message
    level: level to report at. One of: 'critical', 'error', 'warning', 'info', 'debug'
    request: the request object for the context of the message
    extra_data: dictionary of params to include with the message. 'body' is reserved.
    payload_data: param names to pass in the 'data' level of the payload; overrides defaults.
    """
    try:
        return _report_message(message, level, request, extra_data, payload_data)
    except Exception as e:
        log.exception("Exception while reporting message to Rollbar. %r", e)


def send_payload(payload, access_token):
    """
    Sends a payload object, (the result of calling _build_payload() + _serialize_payload()).
    Uses the configured handler from SETTINGS['handler']

    Available handlers:
    - 'blocking': calls _send_payload() (which makes an HTTP request) immediately, blocks on it
    - 'thread': starts a single-use thread that will call _send_payload(). returns immediately.
    - 'async': calls _send_payload_async() (which makes an async HTTP request using default async handler)
    - 'agent': writes to a log file to be processed by rollbar-agent
    - 'tornado': calls _send_payload_tornado() (which makes an async HTTP request using tornado's AsyncHTTPClient)
    - 'gae': calls _send_payload_appengine() (which makes a blocking call to Google App Engine)
    - 'twisted': calls _send_payload_twisted() (which makes an async HTTP request using Twisted and Treq)
    - 'httpx': calls _send_payload_httpx() (which makes an async HTTP request using HTTPX)
    - 'thread_pool': uses a pool of worker threads to make HTTP requests off the main thread. Returns immediately.
    """
    payload = events.on_payload(payload)
    if payload is False:
        return

    if sys.version_info >= (3, 6):
        from rollbar.lib._async import get_current_handler
        handler = get_current_handler()
    else:
        handler = SETTINGS.get('handler')

    if handler == 'twisted':
        payload['data']['framework'] = 'twisted'

    payload_str = _serialize_payload(payload)
    if handler == 'blocking':
        _send_payload(payload_str, access_token)
    elif handler == 'agent':
        agent_log.error(payload_str)
    elif handler == 'tornado':
        if TornadoAsyncHTTPClient is None:
            log.error('Unable to find tornado')
            return
        _send_payload_tornado(payload_str, access_token)
    elif handler == 'gae':
        if AppEngineFetch is None:
            log.error('Unable to find AppEngine URLFetch module')
            return
        _send_payload_appengine(payload_str, access_token)
    elif handler == 'twisted':
        if treq is None:
            log.error('Unable to find Treq')
            return
        _send_payload_twisted(payload_str, access_token)
    elif handler == 'httpx':
        if httpx is None:
            log.error('Unable to find HTTPX')
            return
        _send_payload_httpx(payload_str, access_token)
    elif handler == 'async':
        if AsyncHTTPClient is None:
            log.error('Unable to find async handler')
            return
        _send_payload_async(payload_str, access_token)
    elif handler == 'thread':
        _send_payload_thread(payload_str, access_token)
    elif handler == 'thread_pool':
        _send_payload_thread_pool(payload_str, access_token)
    else:
        # default to 'thread'
        _send_payload_thread(payload_str, access_token)


def search_items(title, return_fields=None, access_token=None, endpoint=None, **search_fields):
    """
    Searches a project for items that match the input criteria.

    title: all or part of the item's title to search for.
    return_fields: the fields that should be returned for each item.
            e.g. ['id', 'project_id', 'status'] will return a dict containing
                 only those fields for each item.
    access_token: a project access token. If this is not provided,
                  the one provided to init() will be used instead.
    search_fields: additional fields to include in the search.
            currently supported: status, level, environment
    """
    if not title:
        return []

    if return_fields is not None:
        return_fields = ','.join(return_fields)

    return _get_api('search/',
                    title=title,
                    fields=return_fields,
                    access_token=access_token,
                    endpoint=endpoint,
                    **search_fields)


def wait(f=None):
    _threads.join()
    if f is not None:
        return f()


class ApiException(Exception):
    """
    This exception will be raised if there was a problem decoding the
    response from an API call.
    """
    pass


class ApiError(ApiException):
    """
    This exception will be raised if the API response contains an 'err'
    field, denoting there was a problem fulfilling the api request.
    """
    pass


class Result(object):
    """
    This class encapsulates the response from an API call.
    Usage:

        result = search_items(title='foo', fields=['id'])
        print result.data
    """

    def __init__(self, access_token, path, params, data):
        self.access_token = access_token
        self.path = path
        self.params = params
        self.data = data

    def __str__(self):
        return str(self.data)


class PagedResult(Result):
    """
    This class wraps the response from an API call that responded with
    a page of results.
    Usage:

        result = search_items(title='foo', fields=['id'])
        print 'First page: %d, data: %s' % (result.page, result.data)
        result = result.next_page()
        print 'Second page: %d, data: %s' % (result.page, result.data)
    """
    def __init__(self, access_token, path, page_num, params, data, endpoint=None):
        super(PagedResult, self).__init__(access_token, path, params, data)
        self.page = page_num
        self.endpoint = endpoint

    def next_page(self):
        params = copy.copy(self.params)
        params['page'] = self.page + 1
        return _get_api(self.path, endpoint=self.endpoint, **params)

    def prev_page(self):
        if self.page <= 1:
            return self
        params = copy.copy(self.params)
        params['page'] = self.page - 1
        return _get_api(self.path, endpoint=self.endpoint, **params)


## internal functions


def _resolve_exception_class(idx, filter):
    cls, level = filter
    if isinstance(cls, str):
        # Lazily resolve class name
        parts = cls.split('.')
        module = '.'.join(parts[:-1])
        if module in sys.modules and hasattr(sys.modules[module], parts[-1]):
            cls = getattr(sys.modules[module], parts[-1])
            SETTINGS['exception_level_filters'][idx] = (cls, level)
        else:
            cls = None
    return cls, level


def _filtered_level(exception):
    for i, filter in enumerate(SETTINGS['exception_level_filters']):
        cls, level = _resolve_exception_class(i, filter)
        if cls and isinstance(exception, cls):
            return level

    return None


def _is_ignored(exception):
    return _filtered_level(exception) == 'ignored'


def _create_agent_log():
    """
    Creates .rollbar log file for use with rollbar-agent
    """
    log_file = SETTINGS['agent.log_file']
    if not log_file.endswith('.rollbar'):
        log.error("Provided agent log file does not end with .rollbar, which it must. "
                  "Using default instead.")
        log_file = DEFAULTS['agent.log_file']

    retval = logging.getLogger('rollbar_agent')
    handler = logging.FileHandler(log_file, 'a', 'utf-8')
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    retval.addHandler(handler)
    retval.setLevel(logging.WARNING)
    return retval


def _report_exc_info(exc_info, request, extra_data, payload_data, level=None):
    """
    Called by report_exc_info() wrapper
    """

    if not _check_config():
        return

    filtered_level = _filtered_level(exc_info[1])
    if level is None:
        level = filtered_level

    filtered_exc_info = events.on_exception_info(exc_info,
                                                 request=request,
                                                 extra_data=extra_data,
                                                 payload_data=payload_data,
                                                 level=level)

    if filtered_exc_info is False:
        return

    cls, exc, trace = filtered_exc_info

    data = _build_base_data(request)
    if level is not None:
        data['level'] = level

    # walk the trace chain to collect cause and context exceptions
    trace_chain = _walk_trace_chain(cls, exc, trace)

    extra_trace_data = None
    if len(trace_chain) > 1:
        data['body'] = {
            'trace_chain': trace_chain
        }
        if payload_data and ('body' in payload_data) and ('trace' in payload_data['body']):
            extra_trace_data = payload_data['body']['trace']
            del payload_data['body']['trace']
    else:
        data['body'] = {
            'trace': trace_chain[0]
        }

    if extra_data:
        extra_data = extra_data
        if not isinstance(extra_data, dict):
            extra_data = {'value': extra_data}
        if extra_trace_data:
            extra_data = dict_merge(extra_data, extra_trace_data, silence_errors=True)
        data['custom'] = extra_data
    if extra_trace_data and not extra_data:
        data['custom'] = extra_trace_data

    request = _get_actual_request(request)
    _add_request_data(data, request)
    _add_person_data(data, request)
    _add_lambda_context_data(data)
    data['server'] = _build_server_data()

    if payload_data:
        data = dict_merge(data, payload_data, silence_errors=True)

    payload = _build_payload(data)
    send_payload(payload, payload.get('access_token'))

    return data['uuid']


def _walk_trace_chain(cls, exc, trace):
    trace_chain = [_trace_data(cls, exc, trace)]

    seen_exceptions = {exc}

    while True:
        exc = getattr(exc, '__cause__', None) or getattr(exc, '__context__', None)
        if not exc:
            break
        trace_chain.append(_trace_data(type(exc), exc, getattr(exc, '__traceback__', None)))
        if exc in seen_exceptions:
            break
        seen_exceptions.add(exc)

    return trace_chain


def _trace_data(cls, exc, trace):
    # exception info
    # most recent call last
    raw_frames = traceback.extract_tb(trace)
    frames = [{'filename': f[0], 'lineno': f[1], 'method': f[2], 'code': f[3]} for f in raw_frames]

    trace_data = {
        'frames': frames,
        'exception': {
            'class': getattr(cls, '__name__', cls.__class__.__name__),
            'message': str(exc),
        }
    }

    _add_locals_data(trace_data, (cls, exc, trace))

    return trace_data


def _report_message(message, level, request, extra_data, payload_data):
    """
    Called by report_message() wrapper
    """
    if not _check_config():
        return

    filtered_message = events.on_message(message,
                                         request=request,
                                         extra_data=extra_data,
                                         payload_data=payload_data,
                                         level=level)

    if filtered_message is False:
        return

    data = _build_base_data(request, level=level)

    # message
    data['body'] = {
        'message': {
            'body': filtered_message
        }
    }

    if extra_data:
        extra_data = extra_data
        data['body']['message'].update(extra_data)

    request = _get_actual_request(request)
    _add_request_data(data, request)
    _add_person_data(data, request)
    _add_lambda_context_data(data)
    data['server'] = _build_server_data()

    if payload_data:
        data = dict_merge(data, payload_data, silence_errors=True)

    payload = _build_payload(data)
    send_payload(payload, payload.get('access_token'))

    return data['uuid']


def _check_config():
    if not SETTINGS.get('enabled'):
        log.info("pyrollbar: Not reporting because rollbar is disabled.")
        return False

    # skip access token check for the agent handler
    if SETTINGS.get('handler') == 'agent':
        return True

    # make sure we have an access_token
    if not SETTINGS.get('access_token'):
        log.warning("pyrollbar: No access_token provided. Please configure by calling rollbar.init() with your access token.")
        return False

    return True


def _build_base_data(request, level='error'):
    data = {
        'timestamp': int(time.time()),
        'environment': SETTINGS['environment'],
        'level': level,
        'language': 'python %s' % '.'.join(str(x) for x in sys.version_info[:3]),
        'notifier': SETTINGS['notifier'],
        'uuid': str(uuid.uuid4()),
    }

    if SETTINGS.get('code_version'):
        data['code_version'] = SETTINGS['code_version']

    if BASE_DATA_HOOK:
        BASE_DATA_HOOK(request, data)

    return data


def _add_person_data(data, request):
    try:
        person_data = _build_person_data(request)
    except Exception as e:
        log.exception("Exception while building person data for Rollbar payload: %r", e)
    else:
        if person_data:
            if not SETTINGS['capture_username'] and 'username' in person_data:
                person_data['username'] = None
            if not SETTINGS['capture_email'] and 'email' in person_data:
                person_data['email'] = None
            data['person'] = person_data


def _build_person_data(request):
    """
    Returns a dictionary describing the logged-in user using data from `request`.

    Try request.rollbar_person first, then 'user', then 'user_id'
    """
    if hasattr(request, 'rollbar_person'):
        rollbar_person_prop = request.rollbar_person
        person = rollbar_person_prop() if callable(rollbar_person_prop) else rollbar_person_prop
        if person and isinstance(person, dict):
            return person
        else:
            return None

    if StarletteRequest:
        from rollbar.contrib.starlette.requests import hasuser
    else:
        def hasuser(request): return True

    if hasuser(request) and hasattr(request, 'user'):
        user_prop = request.user
        user = user_prop() if callable(user_prop) else user_prop
        if not user:
            return None
        elif isinstance(user, dict):
            return user
        else:
            retval = {}
            if getattr(user, 'id', None):
                retval['id'] = str(user.id)
            elif getattr(user, 'user_id', None):
                retval['id'] = str(user.user_id)

            # id is required, so only include username/email if we have an id
            if retval.get('id'):
                username = getattr(user, 'username', None)
                email = getattr(user, 'email', None)
                retval.update({
                    'username': username,
                    'email': email
                })
            return retval

    if hasattr(request, 'user_id'):
        user_id_prop = request.user_id
        user_id = user_id_prop() if callable(user_id_prop) else user_id_prop
        if not user_id:
            return None
        return {'id': str(user_id)}


def _get_func_from_frame(frame):
    func_name = inspect.getframeinfo(frame).function
    caller = frame.f_back
    if caller:
        func = caller.f_locals.get(func_name,
                                   caller.f_globals.get(func_name))
    else:
        func = None

    return func


def _add_locals_data(trace_data, exc_info):
    if not SETTINGS['locals']['enabled']:
        return

    frames = trace_data['frames']

    cur_tb = exc_info[2]
    frame_num = 0
    num_frames = len(frames)
    while cur_tb:
        cur_frame = frames[frame_num]
        tb_frame = cur_tb.tb_frame
        cur_tb = cur_tb.tb_next

        if not isinstance(tb_frame, types.FrameType):
            # this can happen if the traceback or frame is wrapped in some way,
            # for example by `ExceptionInfo` in
            # https://github.com/celery/billiard/blob/master/billiard/einfo.py
            log.warning('Traceback frame not a types.FrameType. Ignoring.')
            frame_num += 1
            continue

        # Create placeholders for argspec/varargspec/keywordspec/locals
        argspec = None
        varargspec = None
        keywordspec = None
        _locals = {}

        try:
            arginfo = inspect.getargvalues(tb_frame)

            # Optionally fill in locals for this frame
            if arginfo.locals and _check_add_locals(cur_frame, frame_num, num_frames):
                # Get all of the named args
                argspec = arginfo.args

                if arginfo.varargs is not None:
                    varargspec = arginfo.varargs
                    if SETTINGS['locals']['scrub_varargs']:
                        temp_varargs = list(arginfo.locals[varargspec])
                        for i, arg in enumerate(temp_varargs):
                            temp_varargs[i] = REDACT_REF

                        arginfo.locals[varargspec] = tuple(temp_varargs)

                if arginfo.keywords is not None:
                    keywordspec = arginfo.keywords

                _locals.update(arginfo.locals.items())

        except Exception:
            log.exception('Error while extracting arguments from frame. Ignoring.')

        # Finally, serialize each arg/kwarg/local separately so that we only report
        # CircularReferences for each variable, instead of for the entire payload
        # as would be the case if we serialized that payload in one-shot.
        if argspec:
            cur_frame['argspec'] = argspec
        if varargspec:
            cur_frame['varargspec'] = varargspec
        if keywordspec:
            cur_frame['keywordspec'] = keywordspec
        if _locals:
            try:
                cur_frame['locals'] = {k: _serialize_frame_data(v) for k, v in _locals.items()}
            except Exception:
                log.exception('Error while serializing frame data.')

        frame_num += 1


def _serialize_frame_data(data):
    return transforms.transform(
        data,
        [ScrubRedactTransform(), _serialize_transform],
        batch_transforms=SETTINGS['batch_transforms']
    )


def _add_lambda_context_data(data):
    """
    Attempts to add information from the lambda context if it exists
    """
    global _CURRENT_LAMBDA_CONTEXT
    context = _CURRENT_LAMBDA_CONTEXT
    if context is None:
        return
    try:
        lambda_data = {
            'lambda': {
                'remaining_time_in_millis': context.get_remaining_time_in_millis(),
                'function_name': context.function_name,
                'function_version': context.function_version,
                'arn': context.invoked_function_arn,
                'request_id': context.aws_request_id,
            }
        }
        if 'custom' in data:
            data['custom'] = dict_merge(data['custom'], lambda_data, silence_errors=True)
        else:
            data['custom'] = lambda_data
    except Exception as e:
        log.exception("Exception while adding lambda context data: %r", e)
    finally:
        _CURRENT_LAMBDA_CONTEXT = None


def _add_request_data(data, request):
    """
    Attempts to build request data; if successful, sets the 'request' key on `data`.
    """
    try:
        request_data = _build_request_data(request)
    except Exception as e:
        log.exception("Exception while building request_data for Rollbar payload: %r", e)
    else:
        if request_data:
            _filter_ip(request_data, SETTINGS['capture_ip'])
            data['request'] = request_data


def _check_add_locals(frame, frame_num, total_frames):
    """
    Returns True if we should record local variables for the given frame.
    """
    # Include the last frames locals
    # Include any frame locals that came from a file in the project's root
    return any(((frame_num == total_frames - 1),
                ('root' in SETTINGS and (frame.get('filename') or '').lower().startswith((SETTINGS['root'] or '').lower()))))


def _get_actual_request(request):
    if WerkzeugLocalProxy and isinstance(request, WerkzeugLocalProxy):
        try:
            actual_request = request._get_current_object()
        except RuntimeError:
            return None
        return actual_request
    return request


def _build_request_data(request):
    """
    Returns a dictionary containing data from the request.
    """

    # webob (pyramid)
    if WebobBaseRequest and isinstance(request, WebobBaseRequest):
        return _build_webob_request_data(request)

    # django
    if DjangoHttpRequest and isinstance(request, DjangoHttpRequest):
        return _build_django_request_data(request)

    # django rest framework
    if RestFrameworkRequest and isinstance(request, RestFrameworkRequest):
        return _build_django_request_data(request)

    # werkzeug (flask)
    if WerkzeugRequest and isinstance(request, WerkzeugRequest):
        return _build_werkzeug_request_data(request)

    # tornado
    if TornadoRequest and isinstance(request, TornadoRequest):
        return _build_tornado_request_data(request)

    # bottle
    if BottleRequest and isinstance(request, BottleRequest):
        return _build_bottle_request_data(request)

    # Sanic
    if SanicRequest and isinstance(request, SanicRequest):
        return _build_sanic_request_data(request)

    # falcon
    if FalconRequest and isinstance(request, FalconRequest):
        return _build_falcon_request_data(request)

    # Plain wsgi (should be last)
    if isinstance(request, dict) and 'wsgi.version' in request:
        return _build_wsgi_request_data(request)

    # FastAPI (built on top of Starlette, so keep the order)
    if FastAPIRequest and isinstance(request, FastAPIRequest):
        return _build_fastapi_request_data(request)

    # Starlette (should be the last one for Starlette based frameworks)
    if StarletteRequest and isinstance(request, StarletteRequest):
        return _build_starlette_request_data(request)

    return None


def _build_webob_request_data(request):
    request_data = {
        'url': request.url,
        'GET': dict(request.GET),
        'user_ip': _extract_user_ip(request),
        'headers': dict(request.headers),
        'method': request.method,
    }

    try:
        if request.json:
            request_data['json'] = request.json
    except:
        pass

    # pyramid matchdict
    if getattr(request, 'matchdict', None):
        request_data['params'] = request.matchdict

    # workaround for webob bug when the request body contains binary data but has a text
    # content-type
    try:
        request_data['POST'] = dict(request.POST)
    except UnicodeDecodeError:
        request_data['body'] = request.body

    return request_data


def _extract_wsgi_headers(items):
    headers = {}
    for k, v in items:
        if k.startswith('HTTP_'):
            header_name = '-'.join(k[len('HTTP_'):].replace('_', ' ').title().split(' '))
            headers[header_name] = v
    return headers


def _build_django_request_data(request):
    try:
        url = request.get_raw_uri()
    except AttributeError:
        url = request.build_absolute_uri()

    request_data = {
        'url': url,
        'method': request.method,
        'GET': dict(request.GET),
        'POST': dict(request.POST),
        'user_ip': _wsgi_extract_user_ip(request.META),
    }

    if SETTINGS['include_request_body']:
        try:
            request_data['body'] = request.body
        except:
            pass

    request_data['headers'] = _extract_wsgi_headers(request.META.items())

    return request_data


def _build_werkzeug_request_data(request):
    request_data = {
        'url': request.url,
        'GET': dict(request.args),
        'POST': dict(request.form),
        'user_ip': _extract_user_ip(request),
        'headers': dict(request.headers),
        'method': request.method,
        'files_keys': list(request.files.keys()),
    }

    try:
        if request.json:
            request_data['body'] = request.json
    except Exception:
        pass

    return request_data


def _build_tornado_request_data(request):
    request_data = {
        'url': request.full_url(),
        'user_ip': request.remote_ip,
        'headers': dict(request.headers),
        'method': request.method,
        'files_keys': request.files.keys(),
        'start_time': getattr(request, '_start_time', None),
    }
    request_data[request.method] = request.arguments

    return request_data


def _build_bottle_request_data(request):
    request_data = {
        'url': request.url,
        'user_ip': request.remote_addr,
        'headers': dict(request.headers),
        'method': request.method,
        'GET': dict(request.query)
    }

    if request.json:
        try:
            request_data['body'] = request.body.getvalue()
        except:
            pass
    else:
        request_data['POST'] = dict(request.forms)

    return request_data


def _build_sanic_request_data(request):
    request_data = {
        'url': request.url,
        'user_ip': request.remote_addr,
        'headers': request.headers,
        'method': request.method,
        'GET': dict(request.args)
    }

    if request.json:
        try:
            request_data['body'] = request.json
        except:
            pass
    else:
        request_data['POST'] = request.form

    return request_data


def _build_falcon_request_data(request):
    request_data = {
        'url': request.url,
        'user_ip': _wsgi_extract_user_ip(request.env),
        'headers': dict(request.headers),
        'method': request.method,
        'GET': dict(request.params),
        'context': dict(request.context),
    }

    return request_data


def _build_wsgi_request_data(request):
    request_data = {
        'url': wsgiref.util.request_uri(request),
        'user_ip': _wsgi_extract_user_ip(request),
        'method': request.get('REQUEST_METHOD'),
    }
    if 'QUERY_STRING' in request:
        request_data['GET'] = parse_qs(request['QUERY_STRING'], keep_blank_values=True)
        # Collapse single item arrays
        request_data['GET'] = {k: (v[0] if len(v) == 1 else v) for k, v in request_data['GET'].items()}

    request_data['headers'] = _extract_wsgi_headers(request.items())

    try:
        length = int(request.get('CONTENT_LENGTH', 0))
    except ValueError:
        length = 0
    input = request.get('wsgi.input')
    if length and input and hasattr(input, 'seek') and hasattr(input, 'tell'):
        pos = input.tell()
        input.seek(0, 0)
        request_data['body'] = input.read(length)
        input.seek(pos, 0)

    return request_data

def _build_starlette_request_data(request):
    from starlette.datastructures import UploadFile

    request_data = {
        'url': str(request.url),
        'GET': dict(request.query_params),
        'headers': dict(request.headers),
        'method': request.method,
        'user_ip': _starlette_extract_user_ip(request),
        'params': dict(request.path_params),
    }

    if hasattr(request, '_form'):
        request_data['POST'] = {
            k: v.filename if isinstance(v, UploadFile) else v
            for k, v in request._form.items()
        }
        request_data['files_keys'] = [
            field.filename
            for field in request._form.values()
            if isinstance(field, UploadFile)
        ]

    if hasattr(request, '_body'):
        body = request._body.decode()
    else:
        body = None

    if body and SETTINGS['include_request_body']:
        request_data['body'] = body

    if hasattr(request, '_json'):
        request_data['json'] = request._json
    elif body:
        try:
            request_data['json'] = json.loads(body)
        except json.JSONDecodeError:
            pass

    # Filter out empty values
    request_data = {k: v for k, v in request_data.items() if v}

    return request_data

def _build_fastapi_request_data(request):
    return _build_starlette_request_data(request)


def _filter_ip(request_data, capture_ip):
    if 'user_ip' not in request_data or capture_ip == True:
        return

    current_ip = request_data['user_ip']
    if not current_ip:
        return

    new_ip = current_ip
    if not capture_ip:
        new_ip = None
    elif capture_ip == ANONYMIZE:
        try:
            if '.' in current_ip:
                new_ip = '.'.join(current_ip.split('.')[0:3]) + '.0'
            elif ':' in current_ip:
                parts = current_ip.split(':')
                if len(parts) > 2:
                    terminal = '0000:0000:0000:0000:0000'
                    new_ip = ':'.join(parts[0:3] + [terminal])
            else:
                new_ip = None
        except:
            new_ip = None

    request_data['user_ip'] = new_ip


def _build_server_data():
    """
    Returns a dictionary containing information about the server environment.
    """
    # server environment
    server_data = {
        'host': socket.gethostname(),
        'pid': os.getpid()
    }

    # argv does not always exist in embedded python environments
    argv = getattr(sys, 'argv', None)
    if argv:
         server_data['argv'] = argv

    for key in ['branch', 'root']:
        if SETTINGS.get(key):
            server_data[key] = SETTINGS[key]

    return server_data


def _transform(obj, key=None):
    return transforms.transform(
        obj,
        _transforms,
        key=key,
        batch_transforms=SETTINGS['batch_transforms']
    )


def _build_payload(data):
    """
    Returns the full payload as a string.
    """

    for k, v in data.items():
        data[k] = _transform(v, key=(k,))

    payload = {
        'access_token': SETTINGS['access_token'],
        'data': data
    }

    return payload


def _serialize_payload(payload):
    return json.dumps(payload, default=defaultJSONEncode)


def _send_payload(payload_str, access_token):
    try:
        _post_api('item/', payload_str, access_token=access_token)
    except Exception as e:
        log.exception('Exception while posting item %r', e)
    try:
        _threads.get_nowait()
        _threads.task_done()
    except queue.Empty:
        pass


def _send_payload_thread(payload_str, access_token):
    thread = threading.Thread(target=_send_payload, args=(payload_str, access_token))
    _threads.put(thread)
    thread.start()


def _send_payload_pool(payload_str, access_token):
    try:
        _post_api('item/', payload_str, access_token=access_token)
    except Exception as e:
        log.exception('Exception while posting item %r', e)


def _send_payload_thread_pool(payload_str, access_token):
    from rollbar.lib.thread_pool import submit
    submit(_send_payload_pool, payload_str, access_token)


def _send_payload_appengine(payload_str, access_token):
    try:
        _post_api_appengine('item/', payload_str, access_token=access_token)
    except Exception as e:
        log.exception('Exception while posting item %r', e)


def _post_api_appengine(path, payload_str, access_token=None):
    headers = {'Content-Type': 'application/json'}

    if access_token is not None:
        headers['X-Rollbar-Access-Token'] = access_token

    url = urljoin(SETTINGS['endpoint'], path)
    resp = AppEngineFetch(url,
                          method="POST",
                          payload=payload_str,
                          headers=headers,
                          allow_truncated=False,
                          deadline=SETTINGS.get('timeout', DEFAULT_TIMEOUT),
                          validate_certificate=SETTINGS.get('verify_https', True))

    return _parse_response(path, SETTINGS['access_token'], payload_str, resp)


def _post_api(path, payload_str, access_token=None):
    headers = {'Content-Type': 'application/json'}

    if access_token is not None:
        headers['X-Rollbar-Access-Token'] = access_token

    url = urljoin(SETTINGS['endpoint'], path)
    resp = transport.post(url,
                          data=payload_str,
                          headers=headers,
                          timeout=SETTINGS.get('timeout', DEFAULT_TIMEOUT),
                          verify=SETTINGS.get('verify_https', True),
                          proxy=SETTINGS.get('http_proxy'),
                          proxy_user=SETTINGS.get('http_proxy_user'),
                          proxy_password=SETTINGS.get('http_proxy_password'))

    return _parse_response(path, SETTINGS['access_token'], payload_str, resp)


def _get_api(path, access_token=None, endpoint=None, **params):
    access_token = access_token or SETTINGS['access_token']
    url = urljoin(endpoint or SETTINGS['endpoint'], path)
    params['access_token'] = access_token
    resp = transport.get(url,
                         params=params,
                         verify=SETTINGS.get('verify_https', True),
                         proxy=SETTINGS.get('http_proxy'),
                         proxy_user=SETTINGS.get('http_proxy_user'),
                         proxy_password=SETTINGS.get('http_proxy_password'))
    return _parse_response(path, access_token, params, resp, endpoint=endpoint)


def _send_payload_tornado(payload_str, access_token):
    try:
        _post_api_tornado('item/', payload_str, access_token=access_token)
    except Exception as e:
        log.exception('Exception while posting item %r', e)


def _post_api_tornado(path, payload_str, access_token=None):
    headers = {'Content-Type': 'application/json'}

    if access_token is not None:
        headers['X-Rollbar-Access-Token'] = access_token
    else:
        access_token = SETTINGS['access_token']

    url = urljoin(SETTINGS['endpoint'], path)

    def post_tornado_cb(resp):
        r = requests.Response()
        r._content = resp.body
        r.status_code = resp.code
        r.headers.update(resp.headers)
        try:
            _parse_response(path, access_token, payload_str, r)
        except Exception as e:
            log.exception('Exception while posting item %r', e)

    TornadoAsyncHTTPClient().fetch(url,
                                   callback=post_tornado_cb,
                                   raise_error=False,
                                   body=payload_str,
                                   method='POST',
                                   connect_timeout=SETTINGS.get('timeout', DEFAULT_TIMEOUT),
                                   request_timeout=SETTINGS.get('timeout', DEFAULT_TIMEOUT))


def _send_payload_twisted(payload_str, access_token):
    try:
        _post_api_twisted('item/', payload_str, access_token=access_token)
    except Exception as e:
        log.exception('Exception while posting item %r', e)

def _post_api_twisted(path, payload_str, access_token=None):
    def post_data_cb(data, resp):
        resp._content = data
        _parse_response(path, SETTINGS['access_token'], payload_str, resp)

    def post_cb(resp):
        r = requests.Response()
        r.status_code = resp.code
        r.headers.update(resp.headers.getAllRawHeaders())
        return treq.content(resp).addCallback(post_data_cb, r)

    headers = {'Content-Type': ['application/json; charset=utf-8']}
    if access_token is not None:
        headers['X-Rollbar-Access-Token'] = [access_token]

    url = urljoin(SETTINGS['endpoint'], path)
    try:
        encoded_payload = payload_str.encode('utf8')
    except (UnicodeDecodeError, UnicodeEncodeError):
        encoded_payload = payload_str

    treq_client = treq.client.HTTPClient(Agent(reactor, contextFactory=VerifyHTTPS()))
    d = treq_client.post(url, encoded_payload, headers=headers,
                  timeout=SETTINGS.get('timeout', DEFAULT_TIMEOUT))
    d.addCallback(post_cb)

def _send_payload_httpx(payload_str, access_token):
    from rollbar.lib._async import call_later, _post_api_httpx
    try:
        call_later(_post_api_httpx('item/', payload_str,
                                   access_token=access_token))
    except Exception as e:
        log.exception('Exception while posting item %r', e)



def _send_payload_async(payload_str, access_token):
    try:
        _send_payload_httpx(payload_str, access_token=access_token)
    except Exception as e:
        log.exception('Exception while posting item %r', e)


def _send_failsafe(message, uuid, host):
    body_message = ('Failsafe from pyrollbar: {0}. Original payload may be found '
                    'in your server logs by searching for the UUID.').format(message)

    data = {
        'level': 'error',
        'environment': SETTINGS['environment'],
        'body': {
            'message': {
                'body': body_message
            }
        },
        'notifier': SETTINGS['notifier'],
        'custom': {
            'orig_uuid': uuid,
            'orig_host': host
        },
        'failsafe': True,
        'internal': True,
    }

    payload = _build_payload(data)

    try:
        send_payload(payload, SETTINGS['access_token'])
    except Exception:
        log.exception('Rollbar: Error sending failsafe.')


def _parse_response(path, access_token, params, resp, endpoint=None):
    if isinstance(resp, requests.Response):
        try:
            data = resp.text
        except Exception:
            data = resp.content
            log.error('resp.text is undefined, resp.content is %r', resp.content)
    else:
        data = resp.content

    global _LAST_RESPONSE_STATUS
    last_response_was_429 = _LAST_RESPONSE_STATUS == 429
    _LAST_RESPONSE_STATUS = resp.status_code

    if resp.status_code == 429:
        if SETTINGS['log_all_rate_limited_items'] or not last_response_was_429:
            log.warning("Rollbar: over rate limit, data was dropped. Payload was: %r", params)
        return
    elif resp.status_code == 502:
        log.exception('Rollbar api returned a 502')
        return
    elif resp.status_code == 413:
        uuid = None
        host = None

        try:
            payload = json.loads(params)
            uuid = payload['data']['uuid']
            host = payload['data']['server']['host']
            log.error("Rollbar: request entity too large for UUID %r\n. Payload:\n%r", uuid, payload)
        except (TypeError, ValueError):
            log.exception('Unable to decode JSON for failsafe.')
        except KeyError:
            log.exception('Unable to find payload parameters for failsafe.')

        _send_failsafe('payload too large', uuid, host)
        # TODO: Should we return here?
    elif resp.status_code != 200:
        log.warning("Got unexpected status code from Rollbar api: %s\nResponse:\n%s",
                    resp.status_code, data)
        # TODO: Should we also return here?

    try:
        json_data = json.loads(data)
    except (TypeError, ValueError):
        log.exception('Could not decode Rollbar api response:\n%s', data)
        raise ApiException('Request to %s returned invalid JSON response', path)
    else:
        if json_data.get('err'):
            raise ApiError(json_data.get('message') or 'Unknown error')

        result = json_data.get('result', {})

        if 'page' in result:
            return PagedResult(access_token, path, result['page'], params, result, endpoint=endpoint)
        else:
            return Result(access_token, path, params, result)


def _extract_user_ip_from_headers(request):
    forwarded_for = request.headers.get('X-Forwarded-For')
    if forwarded_for:
        return forwarded_for
    real_ip = request.headers.get('X-Real-Ip')
    if real_ip:
        return real_ip
    return None


def _extract_user_ip(request):
    return _extract_user_ip_from_headers(request) or request.remote_addr


def _wsgi_extract_user_ip(environ):
    forwarded_for = environ.get('HTTP_X_FORWARDED_FOR')
    if forwarded_for:
        return forwarded_for
    real_ip = environ.get('HTTP_X_REAL_IP')
    if real_ip:
        return real_ip
    return environ['REMOTE_ADDR']


def _starlette_extract_user_ip(request):
    return request.client.host or _extract_user_ip_from_headers(request)



EXCEPTION_INFO = 'exception_info'
MESSAGE = 'message'
PAYLOAD = 'payload'

_event_handlers = {
    EXCEPTION_INFO: [],
    MESSAGE: [],
    PAYLOAD: []
}


def _check_type(typ):
    if typ not in _event_handlers:
        raise ValueError('Unknown type: %s. Must be one of %s' % (typ, _event_handlers.keys()))


def _add_handler(typ, handler_fn, pos):
    _check_type(typ)

    pos = pos if pos is not None else -1
    handlers = _event_handlers[typ]

    try:
        handlers.index(handler_fn)
    except ValueError:
        handlers.insert(pos, handler_fn)


def _remove_handler(typ, handler_fn):
    _check_type(typ)

    handlers = _event_handlers[typ]

    try:
        index = handlers.index(handler_fn)
        handlers.pop(index)
    except ValueError:
        pass


def _on_event(typ, target, **kw):
    _check_type(typ)

    ref = target
    for handler in _event_handlers[typ]:
        result = handler(ref, **kw)
        if result is False:
            return False

        ref = result

    return ref


# Add/remove event handlers

def add_exception_info_handler(handler_fn, pos=None):
    _add_handler(EXCEPTION_INFO, handler_fn, pos)


def remove_exception_info_handler(handler_fn):
    _remove_handler(EXCEPTION_INFO, handler_fn)


def add_message_handler(handler_fn, pos=None):
    _add_handler(MESSAGE, handler_fn, pos)


def remove_message_handler(handler_fn):
    _remove_handler(MESSAGE, handler_fn)


def add_payload_handler(handler_fn, pos=None):
    _add_handler(PAYLOAD, handler_fn, pos)


def remove_payload_handler(handler_fn):
    _remove_handler(PAYLOAD, handler_fn)


# Event handler processing

def on_exception_info(exc_info, **kw):
    return _on_event(EXCEPTION_INFO, exc_info, **kw)


def on_message(message, **kw):
    return _on_event(MESSAGE, message, **kw)


def on_payload(payload, **kw):
    return _on_event(PAYLOAD, payload, **kw)


# Misc

def reset():
    for handlers in _event_handlers.values():
        del handlers[:]
